from __future__ import absolute_import
import weakref
import unicodedata
from fontTools.agl import AGL2UV
from defcon.tools import unicodeTools
from defcon.objects.base import BaseDictObject


class UnicodeData(BaseDictObject):

    """
    This object serves Unicode data for the font.

    **This object posts the following notifications:**

    - UnicodeData.Changed

    This object behaves like a dict. The keys are Unicode values and the
    values are lists of glyph names associated with that unicode value::

        {
            65 : ["A"],
            66 : ["B"],
        }

    To get the list of glyph names associated with a particular Unicode
    value, do this::

        glyphList = unicodeData[65]

    The object defines many more convenient ways of interacting
    with this data.

    .. warning::

        Setting data into this object manually is *highly* discouraged.
        The object automatically keeps itself in sync with the font and the
        glyphs contained in the font. No manual intervention is required.
    """

    changeNotificationName = "UnicodeData.Changed"
    representationFactories = {}

    def __init__(self, layer=None):
        self._layer = None
        if layer is not None:
            self._layer = weakref.ref(layer)
        super(UnicodeData, self).__init__()
        self.beginSelfNotificationObservation()
        self._glyphNameToForcedUnicode = {}
        self._forcedUnicodeToGlyphName = {}

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.layer

    def _get_font(self):
        layerSet = self.layerSet
        if layerSet is None:
            return None
        return layerSet.font

    font = property(_get_font, doc="The :class:`Font` that this object belongs to.")

    def _get_layerSet(self):
        layer = self.layer
        if layer is None:
            return None
        return layer.layerSet

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this object belongs to.")

    def _get_layer(self):
        if self._layer is None:
            return None
        return self._layer()

    layer = property(_get_layer, doc="The :class:`Layer` that this object belongs to.")

    # -----------
    # set and get
    # -----------

    def removeGlyphData(self, glyphName, values):
        """
        Remove the data for the glyph with **glyphName** and
        the Unicode values **values**.

        This should never be called directly.
        """
        for value in values:
            if value not in self:
                continue
            glyphList = self[value]
            if glyphName in glyphList:
                glyphList.remove(glyphName)
            if not glyphList:
                super(UnicodeData, self).__delitem__(value)
        # remove the forced reference to the glyph
        if glyphName in self._glyphNameToForcedUnicode:
            fourcedValue = self._glyphNameToForcedUnicode[glyphName]
            del self._glyphNameToForcedUnicode[glyphName]
            del self._forcedUnicodeToGlyphName[fourcedValue]
        self.postNotification(notification=self.changeNotificationName)

    def addGlyphData(self, glyphName, values):
        """
        Add the data for the glyph with **glyphName** and
        the Unicode values **values**.

        This should never be called directly.
        """
        for value in values:
            # update unicode to glyph name
            glyphList = self.get(value, [])
            if glyphName not in glyphList:
                glyphList.append(glyphName)
            super(UnicodeData, self).__setitem__(value, glyphList)
        self.postNotification(notification=self.changeNotificationName)

    def __delitem__(self, value):
        glyphList = self.get(value)
        if glyphList is None:
            return
        for glyphName in glyphList:
            # remove forced references
            if glyphName in self._glyphNameToForcedUnicode:
                forcedValue = self._glyphNameToForcedUnicode[glyphName]
                del self._forcedUnicodeToGlyphName[forcedValue]
                del self._glyphNameToForcedUnicode[glyphName]
        super(UnicodeData, self).__delitem__(value)
        self.postNotification(notification=self.changeNotificationName)

    def __setitem__(self, value, glyphList):
        if value not in self:
            super(UnicodeData, self).__setitem__(value, [])
        for glyphName in glyphList:
            self[value].append(glyphName)
            # remove now out dated forced references
            if glyphName in self._glyphNameToForcedUnicode:
                forcedValue = self._glyphNameToForcedUnicode[glyphName]
                del self._forcedUnicodeToGlyphName[forcedValue]
                del self._glyphNameToForcedUnicode[glyphName]
        self.postNotification(notification=self.changeNotificationName)

    def clear(self):
        """
        Completely remove all stored data.

        This should never be called directly.
        """
        super(UnicodeData, self).clear()
        self._forcedUnicodeToGlyphName.clear()
        self._glyphNameToForcedUnicode.clear()

    def update(self, other):
        """
        Update the data int this object with the data from **other**.

        This should never be called directly.
        """
        for value, glyphList in other.items():
            for glyphName in glyphList:
                if glyphName in self._glyphNameToForcedUnicode:
                    forcedValue = self._glyphNameToForcedUnicode[glyphName]
                    del self._forcedUnicodeToGlyphName[forcedValue]
                    del self._glyphNameToForcedUnicode[glyphName]
            super(UnicodeData, self).__setitem__(value, list(glyphList))
        self.postNotification(notification=self.changeNotificationName)

    # -------
    # Loaders
    # -------

    def _setupForcedValueDict(self):
        for value, glyphList in self.values():
            if not glyphList:
                glyphName = None
            else:
                glyphName = glyphList[0]
            if value >= _privateUse1Min and value <= _privateUse1Max:
                self._forcedUnicodeToGlyphName[value] = glyphName
            elif value >= _privateUse2Min and value <= _privateUse2Max:
                self._forcedUnicodeToGlyphName[value] = glyphName
            elif value >= _privateUse3Min  and value <= _privateUse3Max:
                self._forcedUnicodeToGlyphName[value] = glyphName

    def _loadForcedUnicodeValue(self, glyphName):
        # already loaded
        if glyphName in self._glyphNameToForcedUnicode:
            return
        # glyph has a real unicode
        if self.unicodeForGlyphName(glyphName) is not None:
            return
        # start at the highest point, falling back to the bottom of the PUA
        startPoint = max(list(self._forcedUnicodeToGlyphName.keys()) + [_privateUse1Min])
        # find the value and store it
        value = _findAvailablePUACode(self._forcedUnicodeToGlyphName)
        self._forcedUnicodeToGlyphName[value] = glyphName
        self._glyphNameToForcedUnicode[glyphName] = value

    # ---------------
    # Value Retrieval
    # ---------------

    def unicodeForGlyphName(self, glyphName):
        """
        Get the Unicode value for **glyphName**. Returns *None*
        if no value is found.
        """
        font = self.font
        if glyphName not in font:
            return None
        glyph = font[glyphName]
        unicodes = glyph.unicodes
        if not unicodes:
            return None
        return unicodes[0]

    def glyphNameForUnicode(self, value):
        """
        Get the first glyph assigned to the Unicode specified
        as **value**. This will return *None* if no glyph is found.
        """
        glyphList = self.get(value)
        if not glyphList:
            return None
        return glyphList[0]

    def pseudoUnicodeForGlyphName(self, glyphName):
        """
        Get the pseudo-Unicode value for **glyphName**.
        This will return *None* if nothing is found.
        """
        realValue = self.unicodeForGlyphName(glyphName)
        if realValue is not None:
            return realValue
        # glyph doesn't have a suffix
        if glyphName.startswith(".") or glyphName.startswith("_"):
            return None
        if "." not in glyphName and "_" not in glyphName:
            return None
        # get the base
        base = glyphName.split(".")[0]
        # in the case of ligatures, grab the first glyph
        base = base.split("_")[0]
        # get the value for the base
        return self.unicodeForGlyphName(base)

    def forcedUnicodeForGlyphName(self, glyphName):
        """
        Get the forced-Unicode value for **glyphName**.
        """
        realValue = self.unicodeForGlyphName(glyphName)
        if realValue is not None:
            return realValue
        if glyphName not in self._glyphNameToForcedUnicode:
            self._loadForcedUnicodeValue(glyphName)
        return self._glyphNameToForcedUnicode[glyphName]

    def glyphNameForForcedUnicode(self, value):
        """
        Get the glyph name assigned to the forced-Unicode
        specified by **value**.
        """
        if value in self:
            glyphName = self[value]
            if isinstance(glyphName, list):
                glyphName = glyphName[0]
            return glyphName
        # A value will not be considered valid until it has
        # been mapped to a glyph name. Therefore, unknown
        # values should return None
        if value not in self._forcedUnicodeToGlyphName:
            return None
        return self._forcedUnicodeToGlyphName[value]

    # ---------------------
    # Description Retrieval
    # ---------------------

    def scriptForGlyphName(self, glyphName, allowPseudoUnicode=True):
        """
        Get the script for **glyphName**. If **allowPseudoUnicode** is
        True, a pseudo-Unicode value will be used if needed. This will
        return *None* if nothing can be found.
        """
        if allowPseudoUnicode:
            value = self.pseudoUnicodeForGlyphName(glyphName)
        else:
            value = self.unicodeForGlyphName(glyphName)
        if value is None:
            return "Unknown"
        return unicodeTools.script(value)

    def blockForGlyphName(self, glyphName, allowPseudoUnicode=True):
        """
        Get the block for **glyphName**. If **allowPseudoUnicode** is
        True, a pseudo-Unicode value will be used if needed. This will
        return *None* if nothing can be found.
        """
        if allowPseudoUnicode:
            value = self.pseudoUnicodeForGlyphName(glyphName)
        else:
            value = self.unicodeForGlyphName(glyphName)
        if value is None:
            return "No_Block"
        return unicodeTools.block(value)

    def categoryForGlyphName(self, glyphName, allowPseudoUnicode=True):
        """
        Get the category for **glyphName**. If **allowPseudoUnicode** is
        True, a pseudo-Unicode value will be used if needed. This will
        return *None* if nothing can be found.
        """
        if allowPseudoUnicode:
            value = self.pseudoUnicodeForGlyphName(glyphName)
        else:
            value = self.unicodeForGlyphName(glyphName)
        if value is None:
            return "Cn"
        return unicodeTools.category(value)

    def decompositionBaseForGlyphName(self, glyphName, allowPseudoUnicode=True):
        """
        Get the decomposition base for **glyphName**. If **allowPseudoUnicode**
        is True, a pseudo-Unicode value will be used if needed. This will
        return *glyphName* if nothing can be found.
        """
        if allowPseudoUnicode:
            uniValue = self.pseudoUnicodeForGlyphName(glyphName)
        else:
            uniValue = self.unicodeForGlyphName(glyphName)
        if uniValue is None:
            return glyphName
        if uniValue is not None:
            font = self.font
            decomposition = unicodeTools.decompositionBase(uniValue)
            if decomposition != -1:
                if decomposition in font.unicodeData:
                    baseGlyphName = font.unicodeData[decomposition][0]
                    if "." in glyphName:
                        suffix = glyphName.split(".", 1)[1]
                        baseWithSuffix = baseGlyphName + "." + suffix
                        if baseWithSuffix in font:
                            baseGlyphName = baseWithSuffix
                    return baseGlyphName
        return glyphName

    def closeRelativeForGlyphName(self, glyphName, allowPseudoUnicode=True):
        """
        Get the close relative for **glyphName**. For example, if you
        request the close relative of the glyph name for the character (,
        you will be given the glyph name for the character ) if it exists
        in the font. If **allowPseudoUnicode** is True, a pseudo-Unicode
        value will be used if needed. This will return *None* if nothing
        can be found.
        """
        return self._openCloseSearch(glyphName, allowPseudoUnicode, unicodeTools.closeRelative)

    def openRelativeForGlyphName(self, glyphName, allowPseudoUnicode=True):
        """
        Get the open relative for **glyphName**. For example, if you
        request the open relative of the glyph name for the character ),
        you will be given the glyph name for the character ( if it exists
        in the font. If **allowPseudoUnicode** is True, a pseudo-Unicode
        value will be used if needed. This will return *None* if nothing
        can be found.
        """
        return self._openCloseSearch(glyphName, allowPseudoUnicode, unicodeTools.openRelative)

    def _openCloseSearch(self, glyphName, allowPseudoUnicode, lookup):
        if allowPseudoUnicode:
            uniValue = self.pseudoUnicodeForGlyphName(glyphName)
        else:
            uniValue = self.unicodeForGlyphName(glyphName)
        if uniValue is None:
            return
        relativeValue = lookup(uniValue)
        # no defined relative value. return.
        if relativeValue is None:
            return
        # look for a hit on the unicode value.
        # if none exists, return.
        preciseMatch = self.glyphNameForUnicode(relativeValue)
        if preciseMatch is None:
            return
        # pseudo unicode is not allowed. use precise match.
        if not allowPseudoUnicode:
            return preciseMatch
        # add the suffix from the provided glyph name to the
        # recise match and test for existence. if it does
        # exist, return it. otherwise fallback to the
        # precise match.
        if "." in glyphName:
            suffix = glyphName.split(".", 1)[1]
            pseudoMatch = preciseMatch + "." + suffix
            if pseudoMatch in self.font:
                return pseudoMatch
        return preciseMatch

    # -------
    # Sorting
    # -------

    def sortGlyphNames(self, glyphNames, sortDescriptors=[dict(type="unicode")]):
        """
        This sorts the list of **glyphNames** following the sort descriptors
        provided in the **sortDescriptors** list. This works by iterating over
        the sort descriptors and subdividing. For example, if the first
        sort descriptor is a suffix type, internally, the result of the
        sort will look something like this::

            [
                [glyphsWithNoSuffix],
                [glyphsWith.suffix1],
                [glyphsWith.suffix2]
            ]

        When the second sort descriptor is processed, the results of previous
        sorts are subdivided even further. For example, if the second
        sort type is script::

            [[
                [glyphsWithNoSuffix, script1], [glyphsWithNoSuffix, script2],
                [glyphsWith.suffix1, script1], [glyphsWith.suffix1, script2],
                [glyphsWith.suffix2, script1], [glyphsWith.suffix2, script2]
            ]]

        And so on. The returned list will be flattened into a list of glyph names.

        Each item in **sortDescriptors** should be a dict of the following form:

        ==================  ===========
        Key                 Description
        ==================  ===========
        type                The type of sort to perform. See below for options.
        ascending           Boolean representing if the glyphs should be in
                            ascending or descending order. Optional. The default is True.
        allowPseudoUnicode  Boolean representing if pseudo-Unicode
                            values are used. If not, real Unicode values will be used
                            if necessary. Optional. The default is False.
        function            A function. Used only for **custom** sort types. See details below.
        ==================  ===========

        *Available Sort Types:*

        There are four types of sort types: simple, complex, canned and custom.
        Simple sorts are based on sorting non-magical values, such as Unicode values.
        Complex sorts are heuristic based sorts based on common glyph name practices,
        aesthetic preferences and other hard to quantify ideas. Custom sorts are just
        that, custom sorts. Canned sorts are combinations of simple, complex and custom
        sorts that give optimized ordering results. Complex and canned sorts may change
        with further updates, so they should not be relied on for persistent ordering.

        ==================   ==============================================================
        Simple Sort Types    Description
        ==================   ==============================================================
        alphabetical         Self-explanatory.
        unicode              Sort based on Unicode value.
        script               Sort based on Unicode script.
        category             Sort based on Unicode category.
        block                Sort based on Unicode block.
        suffix               Sort based on glyph name suffix.
        decompositionBase    Sort based on the base glyph defined in the decomposition rules.
        ==================   ==============================================================

        ==================   ==============================================================
        Complex Sort Types   Description
        ==================   ==============================================================
        weightedSuffix       Sort based on glyph names suffix. The ordering of the
                             suffixes is based on the calculated "weight" of the suffix.
                             This value is calculated by noting what type of glyphs have
                             the same suffix. The more glyph types, the more important the
                             suffix. Additionally, glyphs with suffixes that have only
                             numerical differences are grouped together. For example,
                             a.alt, a.alt1 and a.alt319 will be grouped together.
        ligature             Sort into to groups: non-ligatures and ligatures.
                             The determination of whether a glyph is a ligature or
                             not is based on the Unicode value, common glyph names
                             or the use of an underscore in the name.
        ==================   ==============================================================

        ==================   ==============================================================
        Canned Sort Types    Description
        ==================   ==============================================================
        cannedDesign         Sort glyphs into a design process friendly order.
        ==================   ==============================================================

        ==================   ==============================================================
        Custom Sort Type     Description
        ==================   ==============================================================
        custom               Sort using a custom function. See details below.
        ==================   ==============================================================

        *Sorting with a custom function:*
        If the builtin sort types don't do exactly what you need, you can use a **custom** sort type
        that contains an arbitrary function that handles sorting externally. This follows the same
        sorting logic as detailed above. The custom sort type can be used in conjunction with the
        builtin sort types.

        The function should follow this form::

            mySortFunction(font, glyphNames, ascending=True, allowPseudoUnicode=False)

        The **ascending** and **allowPseudoUnicode** arguments will be the values defined
        in the sort descriptors.

        The function should return a list of lists of glyph names.

        An example::

            def sortByE(font, glyphNames, ascending=True, allowsPseudoUnicodes=False):
                startsWithE = []
                doesNotStartWithE = []
                for glyphName in glyphNames:
                    if glyphName.startswith("startsWithE"):
                        startsWithE.append(glyphName)
                    else:
                        doesNotStartWithE.append(glyphName)
                return [startsWithE, doesNotStartWithE]
        """
        blocks = [glyphNames]
        typeToMethod = dict(
            # simple
            alphabetical=self._sortByAlphabet,
            unicode=self._sortByUnicode,
            category=self._sortByCategory,
            block=self._sortByBlock,
            script=self._sortByScript,
            suffix=self._sortBySuffix,
            decompositionBase=self._sortByDecompositionBase,
            # complex
            weightedSuffix=self._sortByWeightedSuffix,
            ligature=self._sortByLigature,
            # custom
            custom=self._sortByCustomFunction,
            # canned
            cannedDesign=self._cannedSortDesign,
            # private
            _generalType=self._sortByGeneralType,
            _whitespaceCategory=self._sortByWhitespace,
            _containerPartners=self._sortByContainerPartners,
            _manualGroups=self._sortByManualGroups,
            _notdef=self._sortByNotdef,
        )
        for sortDescriptor in sortDescriptors:
            sortType = sortDescriptor["type"]
            ascending = sortDescriptor.get("ascending", True)
            allowPseudoUnicode = sortDescriptor.get("allowPseudoUnicode", False)
            function = sortDescriptor.get("function", None)
            sortMethod = typeToMethod[sortType]

            newBlocks = []
            for block in blocks:
                sortedBlock = self._sortRecurse(blocks, sortMethod, ascending, allowPseudoUnicode, function)
                newBlocks.append(sortedBlock)
            blocks = newBlocks
        return self._flattenSortResult(blocks)

    def _flattenSortResult(self, result):
        final = []
        for i in result:
            if isinstance(i, list):
                final.extend(self._flattenSortResult(i))
            else:
                final.append(i)
        return final

    def _sortRecurse(self, blocks, sortMethod, ascending, allowPseudoUnicode, function):
        if not blocks:
            return []
        if not isinstance(list(blocks)[0], str):
            sortedBlocks = []
            for block in blocks:
                block = self._sortRecurse(block, sortMethod, ascending, allowPseudoUnicode, function)
                sortedBlocks.append(block)
            return sortedBlocks
        else:
            if sortMethod == self._sortByCustomFunction:
                return sortMethod(blocks, ascending, allowPseudoUnicode, function)
            else:
                return sortMethod(blocks, ascending, allowPseudoUnicode)

    # simple sorts

    def _sortByAlphabet(self, glyphNames, ascending, allowPseudoUnicode):
        result = sorted(glyphNames)
        if not ascending:
            result.reverse()
        return result

    def _sortBySuffix(self, glyphNames, ascending, allowPseudoUnicode):
        suffixToGlyphs = {None : []}
        for glyphName in glyphNames:
            if "." not in glyphName or glyphName.startswith("."):
                suffix = None
            else:
                suffix = glyphName.split(".", 1)[1]
            if suffix not in suffixToGlyphs:
                suffixToGlyphs[suffix] = []
            suffixToGlyphs[suffix].append(glyphName)
        result = []
        result.append(suffixToGlyphs.pop(None))
        for suffix, glyphList in sorted(suffixToGlyphs.items()):
            result.append(glyphList)
        return result

    def _sortByUnicode(self, glyphNames, ascending, allowPseudoUnicode):
        withValue = []
        withoutValue = []
        for glyphName in glyphNames:
            if allowPseudoUnicode:
                value = self.pseudoUnicodeForGlyphName(glyphName)
            else:
                value = self.unicodeForGlyphName(glyphName)
            if value is None:
                withoutValue.append(glyphName)
            else:
                withValue.append((value, glyphName))
        withValue = [glyphName for uniValue, glyphName in sorted(withValue)]
        if not ascending:
            withValue.reverse()
            withoutValue.reverse()
            result = [withoutValue, withValue]
        else:
            result = [withValue, withoutValue]
        return result

    def _sortByScript(self, glyphNames, ascending, allowPseudoUnicode):
        return self._sortByUnicodeLookup(glyphNames, ascending, allowPseudoUnicode, self.scriptForGlyphName, unicodeTools.orderedScripts)

    def _sortByBlock(self, glyphNames, ascending, allowPseudoUnicode):
        return self._sortByUnicodeLookup(glyphNames, ascending, allowPseudoUnicode, self.blockForGlyphName, unicodeTools.orderedBlocks)

    def _sortByCategory(self, glyphNames, ascending, allowPseudoUnicode):
        return self._sortByUnicodeLookup(glyphNames, ascending, allowPseudoUnicode, self.categoryForGlyphName, unicodeTools.orderedCategories)

    def _sortByUnicodeLookup(self, glyphNames, ascending, allowPseudoUnicode, tagLookup, orderedTags):
        tagToGlyphs = {}
        for glyphName in glyphNames:
            tag = tagLookup(glyphName, allowPseudoUnicode)
            if tag not in tagToGlyphs:
                tagToGlyphs[tag] = []
            tagToGlyphs[tag].append(glyphName)
        if not orderedTags:
            orderedTags = sorted(tagToGlyphs.keys())
            if None in orderedTags:
                orderedTags.remove(None)
        sortedResult = []
        for tag in orderedTags + [None]:
            if tag in tagToGlyphs:
                sortedResult.append(tagToGlyphs[tag])
        if not ascending:
            sortedResult.reverse()
        return sortedResult

    def _sortByDecompositionBase(self, glyphNames, ascending, allowPseudoUnicode):
        baseToGlyphNames = {None:[]}
        for glyphName in glyphNames:
            if allowPseudoUnicode:
                value = self.pseudoUnicodeForGlyphName(glyphName)
            else:
                value = self.unicodeForGlyphName(glyphName)
            if value is None:
                base = None
            else:
                base = unicodeTools.decompositionBase(value)
                base = self.glyphNameForUnicode(base)
                # try to add the glyph names suffix to the base.
                # this will handle mapping aacute.alt to a.alt
                # instead of aacute.alt to a.
                if base is not None:
                    if "." in glyphName and not glyphName.startswith("."):
                        suffix = glyphName.split(".")[1]
                        if base + "." + suffix in self.font:
                            base = base + "." + suffix
            if base not in baseToGlyphNames:
                baseToGlyphNames[base] = []
            baseToGlyphNames[base].append(glyphName)
        # get the list of glyphs with no base.
        noBase = baseToGlyphNames.pop(None)
        # find all bases that are not in the overall glyph names list
        missingBase = []
        for base in sorted(baseToGlyphNames):
            if base is None:
                continue
            if base not in noBase:
                missingBase.append(base)
        # work through the found bases
        processedBases = set()
        sortedResult = []
        for base in noBase:
            if base in processedBases:
                continue
            processedBases.add(base)
            # the base could be in the list more than once.
            # if so, add the proper number of instances of the base.
            count = noBase.count(base)
            r = [base for i in range(count)]
            # add the referencing glyphs
            r += baseToGlyphNames.get(base, [])
            sortedResult.append(r)
        # work through the missing bases
        for base in sorted(missingBase):
            sortedResult.append(baseToGlyphNames[base])
        # reverse if necessary
        if not ascending:
            sortedResult.reverse()
        return sortedResult

    # custom sorts

    def _sortByCustomFunction(self, glyphNames, ascending, allowPseudoUnicode, function):
        return function(self.font, glyphNames, ascending, allowPseudoUnicode)

    # complex sorts

    def _sortByWeightedSuffix(self, glyphNames, ascending, allowPseudoUnicode):
        # split into two groups
        noSuffix = []
        suffixed = []
        for glyphName in glyphNames:
            if glyphName.startswith(".") or "." not in glyphName:
                noSuffix.append(glyphName)
            else:
                suffixed.append(glyphName)
        if not suffixed:
            return noSuffix
        # magnetize the suffixes
        allSuffixes = set([glyphName.split(".", 1)[-1] for glyphName in suffixed])
        allSuffixes = sorted(allSuffixes)
        magnets = {}
        while allSuffixes:
            toMatch = toMatchOriginal = allSuffixes.pop(0)
            # split off trailing digits
            # this makes .alt1 match .alt2
            stripped = toMatch
            while stripped:
                c = stripped[-1]
                if c.isdigit():
                    stripped = stripped[:-1]
                else:
                    break
            if stripped:
                toMatch = stripped
            # search for suffixes that start with toMatch
            # and have only numerical differences
            matches = []
            for suffix in allSuffixes:
                if suffix.lower().startswith(toMatch.lower()):
                    if len(suffix) == len(toMatch):
                        matches.append(suffix)
                    else:
                        diffLength = len(suffix) - len(toMatch)
                        misMatch = suffix[:-diffLength]
                        allDigits = False not in [c.isdigit() for c in misMatch]
                        if allDigits:
                            matches.append(suffix)
            for suffix in matches:
                allSuffixes.remove(suffix)
            if not matches:
                toMatch = toMatchOriginal
            # store the result
            magnets[toMatch] = [toMatchOriginal] + matches
        # make a flat mapping of suffix : magnet
        suffixToMagnet = {}
        for magnet, suffixes in magnets.items():
            for suffix in suffixes:
                suffixToMagnet[suffix] = magnet
        # gather data about the suffixes
        suffixMap = {}
        for glyphName in suffixed:
            # get the suffix data dict from the map
            suffix = glyphName.split(".", 1)[-1]
            # get the magnet group
            suffix = suffixToMagnet[suffix]
            if suffix not in suffixMap:
                suffixMap[suffix] = dict(glyphNames=[], letter=0, mark=0, number=0, punctuation=0, symbol=0, space=0)
            suffixData = suffixMap[suffix]
            # add the glyph name
            suffixData["glyphNames"].append(glyphName)
            # update the category flags
            category = self.categoryForGlyphName(glyphName, allowPseudoUnicode=allowPseudoUnicode)
            if category in "Lu Ll Lt":
                suffixData["letter"] = True
            elif category in "Lm Mn Mc Me":
                suffixData["mark"] = True
            elif category in "Nd Nl No":
                suffixData["number"] = True
            elif category in "Pc Pd Pe Pf Pi Po Ps":
                suffixData["punctuation"] = True
            elif category in "Sc Sk Sm So":
                suffixData["symbol"] = True
            elif category == "Zs":
                suffixData["space"] = True
        # rank the suffixes
        sorter = []
        for suffix, suffixData in suffixMap.items():
            sortable = (
                suffixData["letter"],
                suffixData["number"],
                suffixData["symbol"],
                suffixData["punctuation"],
                suffixData["mark"],
                suffixData["space"],
                -len(suffixData["glyphNames"]),
                len(suffix),
                suffix,
                suffixData
            )
            sorter.append(sortable)
        suffixSortedGlyphsNames = [item[-1]["glyphNames"] for item in sorted(sorter)]
        if not ascending:
            for i in suffixSortedGlyphsNames:
                i.reverse()
        sortedResult = [noSuffix] + suffixSortedGlyphsNames
        if not ascending:
            sortedResult.reverse()
        return sortedResult

    def _sortByLigature(self, glyphNames, ascending, allowPseudoUnicode):
        notLigature = []
        ligature = []
        for glyphName in glyphNames:
            base = glyphName.split(".")[0]
            if "_" in base and base not in _notReallyLigatures:
                ligature.append(glyphName)
            elif not allowPseudoUnicode and self.unicodeForGlyphName(glyphName) in _ligatureUniValues:
                ligature.append(glyphName)
            elif allowPseudoUnicode and self.pseudoUnicodeForGlyphName(glyphName) in _ligatureUniValues:
                ligature.append(glyphName)
            elif base in ("ff", "fi", "fl", "ffi", "ffl"):
                ligature.append(glyphName)
            else:
                notLigature.append(glyphName)
        final = [notLigature, ligature]
        if not ascending:
            notLigature.reverse()
            ligature.reverse()
            final.reverse()
        return final

    # canned sorts

    def _cannedSortDesign(self, glyphNames, ascending, allowPseudoUnicode):
        sortDescriptors = [
            dict(type="weightedSuffix", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="ligature", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="_generalType", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="_whitespaceCategory", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="alphabetical", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="unicode", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="category", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="script", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="decompositionBase", allowPseudoUnicode=allowPseudoUnicode),
        ]
        glyphNames = self.sortGlyphNames(glyphNames, sortDescriptors)
        sortDescriptors = [
            dict(type="_containerPartners", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="_manualGroups", allowPseudoUnicode=allowPseudoUnicode),
            dict(type="_notdef", allowPseudoUnicode=allowPseudoUnicode),
        ]
        glyphNames = self.sortGlyphNames(glyphNames, sortDescriptors)
        if not ascending:
            glyphNames.reverse()
        return glyphNames

    # private sorts

    def _sortByGeneralType(self, glyphNames, ascending, allowPseudoUnicode):
        noSuffixUnicode = []
        noSuffixNoUnicode = []
        suffix = []
        for glyphName in glyphNames:
            if glyphName.startswith("."):
                noSuffixNoUnicode.append(glyphName)
            elif "." in glyphName:
                suffix.append(glyphName)
            elif self.unicodeForGlyphName(glyphName):
                noSuffixUnicode.append(glyphName)
            else:
                noSuffixNoUnicode.append(glyphName)
        final = [noSuffixUnicode, noSuffixNoUnicode, suffix]
        if not ascending:
            noSuffixUnicode.reverse()
            noSuffixNoUnicode.reverse()
            suffix.reverse()
            final.reverse
        return final

    def _sortByWhitespace(self, glyphNames, ascending, allowPseudoUnicode):
        space = []
        notSpace = []
        for glyphName in glyphNames:
            if self.categoryForGlyphName(glyphName, allowPseudoUnicode=allowPseudoUnicode) == "Zs":
                space.append(glyphName)
            else:
                notSpace.append(glyphName)
        final = [space, notSpace]
        if not ascending:
            space.reverse()
            notSpace.reverse()
            final.reverse()
        return final

    def _sortByContainerPartners(self, glyphNames, ascending, allowPseudoUnicode):
        glyphOrder = []
        while glyphNames:
            glyphName = glyphNames[0]
            glyphOrder.append(glyphName)
            glyphNames = glyphNames[1:]
            # get the close relative
            closeRelative = self.closeRelativeForGlyphName(glyphName, allowPseudoUnicode=allowPseudoUnicode)
            if closeRelative is not None:
                if closeRelative not in glyphOrder:
                    glyphOrder.append(closeRelative)
                if closeRelative in glyphNames:
                    glyphNames.remove(closeRelative)
        if not ascending:
            glyphOrder.reverse()
        return glyphOrder

    def _sortByNotdef(self, glyphNames, ascending, allowPseudoUnicode):
        notDef = []
        notNotDef = []
        for glyphName in glyphNames:
            if glyphName.startswith(".notdef"):
                notDef.append(glyphName)
            else:
                notNotDef.append(glyphName)
        return notNotDef + notDef

    def _sortByManualGroups(self, glyphNames, ascending, allowPseudoUnicode):
        # build a unicode map for all glyphs
        pseudoUnicodeMapping = {}
        for glyphName in glyphNames:
            if allowPseudoUnicode:
                uniValue = self.pseudoUnicodeForGlyphName(glyphName)
            else:
                uniValue = self.unicodeForGlyphName(glyphName)
            if uniValue not in pseudoUnicodeMapping:
                pseudoUnicodeMapping[uniValue] = []
            pseudoUnicodeMapping[uniValue].append(glyphName)
        for pairGroup in _manualSortGroups:
            # find the matching glyphs
            matched = []
            for uniValue in pairGroup:
                matched += pseudoUnicodeMapping.get(uniValue, [])
            # break the matched into suffix groups
            suffixes = {}
            for glyphName in matched:
                if "." not in glyphName:
                    suffix = None
                else:
                    suffix = glyphName.split(".", 1)[-1]
                if suffix not in suffixes:
                    suffixes[suffix] = []
                suffixes[suffix].append(glyphName)
            # wor through the suffixes
            for matched in suffixes.values():
                # remove matched[1:] them after matched[0]
                if len(matched) > 1:
                    for glyphName in matched[1:]:
                        glyphNames.remove(glyphName)
                    insertionPoint = glyphNames.index(matched[0])
                    glyphNames = glyphNames[:insertionPoint + 1] + matched[1:] + glyphNames[insertionPoint + 1:]
        return glyphNames

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(UnicodeData, self).endSelfNotificationObservation()
        self._layer = None


# -----
# Tools
# -----

# PUA lookups

_privateUse1Min = int("E000", 16)
_privateUse1Max = int("F8FF", 16)
_privateUse2Min = int("F0000", 16)
_privateUse2Max = int("FFFFD", 16)
_privateUse3Min = int("100000", 16)
_privateUse3Max = int("10FFFD", 16)
_viablePUACount = (_privateUse1Max - _privateUse1Min) + (_privateUse2Max - _privateUse2Min) + (_privateUse3Max - _privateUse3Min)

def _findAvailablePUACode(existing, code=None):
    assert len(existing) < _viablePUACount

    if code is None:
        code = _privateUse1Min
    else:
        code += 1

    # force the code into a viable position. this will prevent
    # iteration over values that are not in PUA.
    if code > _privateUse1Max and code < _privateUse2Min:
        code = _privateUse2Min
    elif code > _privateUse2Max and code < _privateUse3Min:
        code = _privateUse3Min
    elif code > _privateUse3Max:
        code = _privateUse1Min

    if code >= _privateUse1Min and code <= _privateUse1Max:
        if code not in existing:
            return code
        else:
            return _findAvailablePUACode(existing, code)
    elif code >= _privateUse2Min and code <= _privateUse2Max:
        if code not in existing:
            return code
        else:
            return _findAvailablePUACode(existing, code)
    else:
        if code < _privateUse3Min or code > _privateUse3Max:
            code = _privateUse3Min
        if code not in existing:
            return code
        else:
            return _findAvailablePUACode(existing, code)

# -----------------
# Sorting Constants
# -----------------

_notReallyLigatures = "space_uni0326".split(" ")
_ligatureUniValues = [int(i, 16) for i in "FB00 FB01 FB02 FB03 FB04 FB05 FB06".split(" ")]

_manualSortGroups = """
percent
perthousand
0x2031

slash
backslash
fraction
bar
brokenbar

period
comma
colon
semicolon
ellipsis
exclam
exclamdown
exclamdbl
question
questiondown

quotesingle
quotedbl
quoteleft
quoteright
quotedblleft
quotedblright
0x201B
0x201F
quotesinglbase
quotedblbase
guilsinglleft
guilsinglright
guillemotleft
guillemotright

at
ampersand
section
paragraph
0x3D7
0x2113
0x2139
0x2116

asterisk
dagger
daggerdbl
0x2042

periodcentered
bullet

acute
grave
hungarumlaut
circumflex
caron
breve
tilde
macron
dieresis
dotaccent
ring
cedilla
ogonek

asciicircum
asciitilde

plus
minus
plusminus
divide
multiply
equal
less
greater
lessequal
greaterequal
approxequal
notequal

copyright
0x24C5 # circle P
registered
trademark
0x2120 # SM

summation
0x00B5

Euro
florin

0x00B9
0x00B2
0x00B3
""".strip()
_currentGroup = []
_parsed = [_currentGroup]
for line in _manualSortGroups.splitlines():
    if "#" in line:
        line = line.split("#")[0].strip()
        assert line
    if not line:
        if _currentGroup:
            _currentGroup = []
            _parsed.append(_currentGroup)
        continue
    if line.startswith("0x"):
        line = int(line[2:], 16)
    else:
        line = AGL2UV[line]
    _currentGroup.append(line)
for g in _parsed:
    assert len(g)
_manualSortGroups = _parsed


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import unittest
from defcon.tools.notifications import NotificationCenter
from defcon.test.testTools import NotificationTestObserver


class _TestObservable(object):

    def __init__(self, center, name):
        self.center = center
        self.name = name

    def __repr__(self):
        return "<_TestObservable {name} {id}".format(name=self.name, id=id(self))

    def postNotification(self, name):
        self.center.postNotification(name, self)


class NotificationCenterTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_addObserver_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        self.assertTrue(center.hasObserver(observer, "A", observable))
        self.assertFalse(center.hasObserver(observer, "B", observable))

    def test_addObserver_notification_no_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", None)
        self.assertTrue(center.hasObserver(observer, "A", None))
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_addObserver_no_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, observable)
        self.assertTrue(center.hasObserver(observer, None, observable))
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_addObserver_no_notification_no_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, None)
        self.assertTrue(center.hasObserver(observer, None, None))
        self.assertFalse(center.hasObserver(observer, "A", observable))
        self.assertFalse(center.hasObserver(observer, None, observable))
        self.assertFalse(center.hasObserver(observer, "A", None))

    def test_removeObserver(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.removeObserver(observer, "A", observable)
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_removeObserver_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.removeObserver(observer, "A", observable)
        self.assertFalse(center.hasObserver(observer, "A", observable))

    def test_removeObserver_allNotifications(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer1 = NotificationTestObserver()
        observer2 = NotificationTestObserver()
        center.addObserver(observer1, "notificationCallback", "A", observable1)
        center.addObserver(observer1, "notificationCallback", "B", observable1)
        center.addObserver(observer2, "notificationCallback", "A", observable1)
        center.addObserver(observer2, "notificationCallback", "B", observable1)
        center.addObserver(observer1, "notificationCallback", "A", observable2)
        center.addObserver(observer1, "notificationCallback", "B", observable2)
        center.addObserver(observer2, "notificationCallback", "A", observable2)
        center.addObserver(observer2, "notificationCallback", "B", observable2)
        center.removeObserver(observer1, "all", observable1)
        self.assertFalse(center.hasObserver(observer1, "A", observable1))
        self.assertFalse(center.hasObserver(observer1, "B", observable1))
        self.assertTrue(center.hasObserver(observer1, "A", observable2))
        self.assertTrue(center.hasObserver(observer1, "B", observable2))
        self.assertTrue(center.hasObserver(observer2, "A", observable1))
        self.assertTrue(center.hasObserver(observer2, "B", observable1))
        self.assertTrue(center.hasObserver(observer2, "A", observable2))
        self.assertTrue(center.hasObserver(observer2, "B", observable2))

    def test_removeObserver_notification_no_observable(self):
        center = NotificationCenter()
        _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", None)
        center.removeObserver(observer, "A", None)
        self.assertFalse(center.hasObserver(observer, "A", None))

    def test_removeObserver_no_notification_observable(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, observable)
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.removeObserver(observer, None, observable)
        self.assertFalse(center.hasObserver(observer, None, observable))
        self.assertTrue(center.hasObserver(observer, "A", observable))

    def test_removeObserver_no_notification_no_observable(self):
        center = NotificationCenter()
        _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, None)
        center.removeObserver(observer, None, None)
        self.assertFalse(center.hasObserver(observer, None, None))

    def test_postNotification_notification_observable(self):
        # notification, observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("B", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_postNotification_notification_no_observable(self):
        # notification, no observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", None)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))
        center.postNotification("B", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))

    def test_postNotification_no_notification_observable(self):
        # no notification, observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, observable1)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        center.postNotification("B", observable2)

    def test_postNotification_no_notification_no_observable(self):
        # no notification, no observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", None, None)
        center.postNotification("A", observable1)
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        center.postNotification("A", observable2)
        self.assertEqual(observer.stack[-1], ("A", "Observable2"))
        center.postNotification("B", observable1)
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        center.postNotification("B", observable2)
        self.assertEqual(observer.stack[-1], ("B", "Observable2"))

    def test_hold_and_releaseHeldNotifications_all_notifications(self):
        "Hold all notifications"
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.holdNotifications()
        observable1.postNotification("A")
        self.assertEqual(len(observer.stack), 0)
        observable1.postNotification("A")
        self.assertEqual(len(observer.stack), 0)
        observable1.postNotification("B")
        self.assertEqual(len(observer.stack), 0)
        observable2.postNotification("C")
        self.assertEqual(len(observer.stack), 0)
        center.releaseHeldNotifications()
        self.assertEqual(observer.stack[-3], ("A", "Observable1"))
        self.assertEqual(observer.stack[-2], ("B", "Observable1"))
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        self.assertEqual(len(observer.stack), 3)

    def test_hold_and_releaseHeldNotifications_notifications_of_observable(self):
        "Hold all notifications of a specific observable"
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.holdNotifications(observable=observable1)
        observable1.postNotification("A")
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.releaseHeldNotifications(observable=observable1)
        self.assertEqual(observer.stack[-2], ("A", "Observable1"))
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        self.assertEqual(len(observer.stack), 3)

    def test_hold_and_releaseHeldNotifications_notifications_of_observable(
            self):
        "Hold all notifications of a specific notification"
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.holdNotifications(notification="A")
        observable1.postNotification("A")
        observable1.postNotification("A")
        observable1.postNotification("B")
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.releaseHeldNotifications(notification="A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))
        self.assertEqual(len(observer.stack), 3)

    def test_areNotificationsHeld_all_held(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.holdNotifications()
        self.assertTrue(center.areNotificationsHeld())
        center.releaseHeldNotifications()
        self.assertFalse(center.areNotificationsHeld())

    def test_areNotificationsHeld_observable_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.holdNotifications(observable=observable1)
        self.assertTrue(center.areNotificationsHeld(observable=observable1))
        self.assertFalse(center.areNotificationsHeld(observable=observable2))
        center.releaseHeldNotifications(observable=observable1)
        self.assertFalse(center.areNotificationsHeld(observable=observable1))

    def test_areNotificationsHeld_notification_off(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable)
        center.addObserver(observer, "notificationCallback", "B", observable)
        center.holdNotifications(notification="A")
        self.assertTrue(center.areNotificationsHeld(notification="A"))
        self.assertFalse(center.areNotificationsHeld(notification="B"))
        center.releaseHeldNotifications(notification="A")
        self.assertFalse(center.areNotificationsHeld(notification="A"))

    def test_areNotificationsHeld_observer_off(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer1 = NotificationTestObserver()
        observer2 = NotificationTestObserver()
        center.addObserver(observer1, "notificationCallback", "A", observable)
        center.addObserver(observer2, "notificationCallback", "A", observable)
        center.holdNotifications(observer=observer1)
        self.assertTrue(center.areNotificationsHeld(observer=observer1))
        self.assertFalse(center.areNotificationsHeld(observer=observer2))
        center.releaseHeldNotifications(observer=observer1)
        self.assertFalse(center.areNotificationsHeld(observer=observer1))

    def test_disable_enableNotifications_all_notifications(self):
        # disable all notifications
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications()
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        center.enableNotifications()
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_disable_enableNotifications_specific_observable(self):
        # disable all notifications for a specific observable
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications(observable=observable1)
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.enableNotifications(observable=observable1)
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_disable_enableNotifications_specific_notification(self):
        # disable all notifications for a specific notification
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications(notification="A")
        observable1.postNotification("A")
        observable1.postNotification("B")
        self.assertEqual(observer.stack[-1], ("B", "Observable1"))
        observable2.postNotification("C")
        self.assertEqual(observer.stack[-1], ("C", "Observable2"))
        center.enableNotifications(notification="A")
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_disable_enableNotifications_specific_observer(self):
        # disable all notifications for a specific observer
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable1)
        center.addObserver(observer, "notificationCallback", "C", observable2)
        center.disableNotifications(observer=observer)
        observable1.postNotification("A")
        observable1.postNotification("B")
        observable2.postNotification("C")
        center.enableNotifications(observer=observer)
        observable1.postNotification("A")
        self.assertEqual(observer.stack[-1], ("A", "Observable1"))

    def test_areNotificationsDisabled_all_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.disableNotifications()
        self.assertTrue(center.areNotificationsDisabled())
        center.enableNotifications()
        self.assertFalse(center.areNotificationsDisabled())

    def test_areNotificationsDisabled_observable_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.disableNotifications(observable=observable1)
        self.assertTrue(
            center.areNotificationsDisabled(observable=observable1))
        self.assertFalse(
            center.areNotificationsDisabled(observable=observable2))
        center.enableNotifications(observable=observable1)
        self.assertFalse(
            center.areNotificationsDisabled(observable=observable1))

    def test_areNotificationsDisabled_notification_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable1)
        center.addObserver(observer, "notificationCallback", "B", observable2)
        center.disableNotifications(notification="A")
        self.assertTrue(center.areNotificationsDisabled(notification="A"))
        self.assertFalse(center.areNotificationsDisabled(notification="B"))
        center.enableNotifications(notification="A")
        self.assertFalse(center.areNotificationsDisabled(notification="A"))

    def test_areNotificationsDisabled_observer_off(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observer1 = NotificationTestObserver()
        observer2 = NotificationTestObserver()
        center.addObserver(observer1, "notificationCallback", "A", observable1)
        center.addObserver(observer2, "notificationCallback", "A", observable1)
        center.disableNotifications(observer=observer1)
        self.assertTrue(center.areNotificationsDisabled(observer=observer1))
        self.assertFalse(center.areNotificationsDisabled(observer=observer2))
        center.enableNotifications(observer=observer1)
        self.assertFalse(center.areNotificationsDisabled(observer=observer1))

    def test_addObserver_identifier(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable, identifier="identifier1")
        center.addObserver(observer, "notificationCallback", "B", observable, identifier="identifier2")
        expected = [
            dict(observer=observer, observable=observable, notification="A", identifier="identifier1"),
            dict(observer=observer, observable=observable, notification="B", identifier="identifier2")
        ]
        result = center.findObservations()
        self.assertEqual(result, expected)

    def test_addObserver_identifierDuplicate(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable, identifier="identifier1")
        center.addObserver(observer, "notificationCallback", "B", observable, identifier="identifier1")
        expected = [
            dict(observer=observer, observable=observable, notification="A", identifier="identifier1"),
            dict(observer=observer, observable=observable, notification="B", identifier="identifier1")
        ]
        self.assertEqual(center.findObservations(), expected)

    def test_removeObserver_identifier(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable, identifier="identifier1")
        center.addObserver(observer, "notificationCallback", "B", observable, identifier="identifier2")
        center.removeObserver(observer, "A", observable)
        center.removeObserver(observer, "B", observable)
        expected = []
        self.assertEqual(center.findObservations(), expected)

    def _buildFindObservationsObjects(self):
        center = NotificationCenter()
        observable1 = _TestObservable(center, "Observable1")
        observable2 = _TestObservable(center, "Observable2")
        observer1 = NotificationTestObserver(name="Observer1")
        observer2 = NotificationTestObserver(name="Observer2")
        center.addObserver(observer1, "notificationCallback", "A", observable1, identifier="identifier1-1-A")
        center.addObserver(observer1, "notificationCallback", "B", observable1, identifier="identifier1-1-B")
        center.addObserver(observer1, "notificationCallback", "A", observable2, identifier="identifier1-2-A")
        center.addObserver(observer1, "notificationCallback", "B", observable2, identifier="identifier1-2-B")
        center.addObserver(observer2, "notificationCallback", "A", observable1, identifier="identifier2-1-A")
        center.addObserver(observer2, "notificationCallback", "B", observable1, identifier="identifier2-1-B")
        center.addObserver(observer2, "notificationCallback", "A", observable2, identifier="identifier2-2-A")
        center.addObserver(observer2, "notificationCallback", "B", observable2, identifier="identifier2-2-B")
        return center, observer1, observer2, observable1, observable2

    def test_findObservations_observer(self):
        center, observer1, observer2, observable1, observable2 = self._buildFindObservationsObjects()
        expected = [
            dict(observer=observer1, notification="A", observable=observable1, identifier="identifier1-1-A"),
            dict(observer=observer1, notification="B", observable=observable1, identifier="identifier1-1-B"),
            dict(observer=observer1, notification="A", observable=observable2, identifier="identifier1-2-A"),
            dict(observer=observer1, notification="B", observable=observable2, identifier="identifier1-2-B")
        ]
        result = center.findObservations(observer=observer1)
        self.assertEqual(result, expected)

    def test_findObservations_observable(self):
        center, observer1, observer2, observable1, observable2 = self._buildFindObservationsObjects()
        expected = [
            dict(observer=observer1, notification="A", observable=observable1, identifier="identifier1-1-A"),
            dict(observer=observer2, notification="A", observable=observable1, identifier="identifier2-1-A"),
            dict(observer=observer1, notification="B", observable=observable1, identifier="identifier1-1-B"),
            dict(observer=observer2, notification="B", observable=observable1, identifier="identifier2-1-B")
        ]
        result = center.findObservations(observable=observable1)
        self.assertEqual(result, expected)

    def test_findObservations_notification(self):
        center, observer1, observer2, observable1, observable2 = self._buildFindObservationsObjects()
        expected = [
            dict(observer=observer1, notification="A", observable=observable1, identifier="identifier1-1-A"),
            dict(observer=observer2, notification="A", observable=observable1, identifier="identifier2-1-A"),
            dict(observer=observer1, notification="A", observable=observable2, identifier="identifier1-2-A"),
            dict(observer=observer2, notification="A", observable=observable2, identifier="identifier2-2-A")
        ]
        result = center.findObservations(notification="A")
        self.assertEqual(result, expected)

    def test_findObservations_identifier(self):
        center, observer1, observer2, observable1, observable2 = self._buildFindObservationsObjects()
        expected = [
            dict(observer=observer1, notification="A", observable=observable1, identifier="identifier1-1-A")
        ]
        result = center.findObservations(identifier="identifier1-1-A")
        self.assertEqual(result, expected)

    def test_findObservations_identifierPattern(self):
        center, observer1, observer2, observable1, observable2 = self._buildFindObservationsObjects()
        expected = [
            dict(observer=observer1, notification="A", observable=observable1, identifier="identifier1-1-A"),
            dict(observer=observer1, notification="A", observable=observable2, identifier="identifier1-2-A")
        ]
        result = center.findObservations(identifier="identifier1-*-A")
        self.assertEqual(result, expected)

    def test_findObservations_all(self):
        center, observer1, observer2, observable1, observable2 = self._buildFindObservationsObjects()
        expected = [
            dict(observer=observer1, notification="A", observable=observable1, identifier="identifier1-1-A")
        ]
        result = center.findObservations(observer=observer1, notification="A", observable=observable1, identifier="identifier1-1-A")
        self.assertEqual(result, expected)

    def test_findObservations_noIdentifier(self):
        center = NotificationCenter()
        observable = _TestObservable(center, "Observable")
        observer = NotificationTestObserver()
        center.addObserver(observer, "notificationCallback", "A", observable, identifier="identifier1")
        center.addObserver(observer, "notificationCallback", "B", observable)
        expected = [
            dict(observer=observer, observable=observable, notification="A", identifier="identifier1"),
            dict(observer=observer, observable=observable, notification="B", identifier=None)
        ]
        result = center.findObservations()
        self.assertEqual(result, expected)

if __name__ == "__main__":
    unittest.main()



"""
A flexible and relatively robust implementation
of the Observer Pattern.
"""

from __future__ import print_function
from collections import OrderedDict
import sys
import weakref
from fnmatch import fnmatchcase

"""

----------------------
Internal Documentation
----------------------

Storage Structures:

registry : {
        (notification, observable) : OrderedDict(
            observer : method name
        )
}

holds : {
    (notification, observable, observer) : {
        count=int,
        notifications=[
            (notification name, observable ref, data)
        ]
    )
}

disabled : {
    (notification, observable, observer) : count
}

identifierRegistry : {
        (notification, observable) : OrderedDict(
            observer : identifier
        )
}

"""


class NotificationCenter(object):

    def __init__(self):
        self._registry = {}
        self._observerKeyBacktrack = {}
        self._holds = {}
        self._disabled = {}
        self._identifierRegistry = {}

    # -----
    # Basic
    # -----

    def addObserver(self, observer, methodName, notification=None, observable=None, identifier=None):
        """
        Add an observer to this notification dispatcher.

        * **observer** An object that can be referenced with weakref.
        * **methodName** A string representing the method to be called
          when the notification is posted.
        * **notification** The notification that the observer should
          be notified of. If this is None, all notifications for
          the *observable* will be posted to *observer*.
        * **observable** The object to observe. If this is None,
          all notifications with the name provided as *notification*
          will be posted to the *observer*.
        * **identifier** None or a string identifying the observation.
          There is no requirement that the string be unique. A reverse
          domain naming scheme is recommended, but there are no
          requirements for the structure of the string.

        If None is given for both *notification* and *observable*
        **all** notifications posted will be sent to the given method
        of the observer.

        The method that will be called as a result of the action
        must accept a single *notification* argument. This will
        be a :class:`Notification` object.
        """
        assert notification != "all", "'all' is a reserved name and can not be used as a notification."
        if observable is not None:
            observable = weakref.ref(observable)
        observer = weakref.ref(observer)
        key = (notification, observable)
        if key not in self._registry:
            self._registry[key] = OrderedDict()
        assert observer not in self._registry[key],\
            "An observer is only allowed to have one callback for a given notification + observable combination. notification={notification}, observable={observable} , observer={observer}, existing method={method1}, adding method={method2}".format(
                notification=key[0], observable=key[1](), observer=observer(), method1=self._registry[key][observer], method2=methodName
            )
        self._registry[key][observer] = methodName
        if observer not in self._observerKeyBacktrack:
            self._observerKeyBacktrack[observer] = {}
        if observable not in self._observerKeyBacktrack[observer]:
            self._observerKeyBacktrack[observer][observable] = set()
        self._observerKeyBacktrack[observer][observable].add(key)
        if identifier is not None:
            if key not in self._identifierRegistry:
                self._identifierRegistry[key] = OrderedDict()
            self._identifierRegistry[key][observer] = identifier

    def hasObserver(self, observer, notification, observable):
        """
        Returns a boolean indicating if the **observer** is registered
        for **notification** posted by **observable**. Either
        *observable* or *notification* may be None.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        key = (notification, observable)
        if key not in self._registry:
            return False
        observer = weakref.ref(observer)
        return observer in self._registry[key]

    def removeObserver(self, observer, notification, observable=None):
        """
        Remove an observer from this notification dispatcher.

        * **observer** A registered object.
        * **notification** The notification that the observer was registered
          to be notified of. If this is `"all"`, all notifications for
          the *observable* will be removed for *observer*.
        * **observable** The object being observed.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        observer = weakref.ref(observer)
        if notification == "all":
            keys = set(self._observerKeyBacktrack[observer][observable])
        else:
            keys = [
                (notification, observable)
            ]
        for key in keys:
            if key not in self._registry:
                continue
            if observer in self._registry[key]:
                del self._registry[key][observer]
            if not len(self._registry[key]):
                del self._registry[key]
            if observer in self._observerKeyBacktrack:
                if observable in self._observerKeyBacktrack[observer]:
                    if key in self._observerKeyBacktrack[observer][observable]:
                        self._observerKeyBacktrack[observer][observable].remove(key)
                    if not self._observerKeyBacktrack[observer][observable]:
                        del self._observerKeyBacktrack[observer][observable]
                if not self._observerKeyBacktrack[observer]:
                    del self._observerKeyBacktrack[observer]
            if key in self._identifierRegistry:
                if observer in self._identifierRegistry[key]:
                    del self._identifierRegistry[key][observer]
                if not len(self._identifierRegistry[key]):
                    del self._identifierRegistry[key]

    def postNotification(self, notification, observable, data=None):
        assert notification is not None
        assert observable is not None
        observableRef = weakref.ref(observable)
        # observer independent hold/disabled
        # ----------------------------------
        if self._holds or self._disabled:
            holdDisabledPossibilities = (
                # least specific -> most specific
                # suspended for all
                (None, None, None),
                # suspended for this notification
                (notification, None, None),
                # suspended for this observer
                (None, observableRef, None),
                # suspended for this notification + observable
                (notification, observableRef, None)
            )
            for key in holdDisabledPossibilities:
                if key in self._disabled:
                    return
                if key in self._holds:
                    n = (notification, observableRef, data)
                    if n not in self._holds[key]["notifications"]:
                        self._holds[key]["notifications"].append(n)
                    return
        # posting
        # -------
        notificationObj = Notification(notification, observableRef, data)
        registryPossibilities = (
            # least specific -> most specific
            (None, None),
            (None, observableRef),
            (notification, None),
            (notification, observableRef),
        )
        for key in registryPossibilities:
            if key not in self._registry:
                continue
            for observerRef, methodName in list(self._registry[key].items()):
                # observer specific hold/disabled
                # -------------------------------
                if self._holds or self._disabled:
                    holdDisabledPossibilities = (
                        # least specific -> most specific
                        # suspended for observer
                        (None, None, observerRef),
                        # suspended for notification + observer
                        (notification, None, observerRef),
                        # suspended for observable + observer
                        (None, observableRef, observerRef),
                        # suspended for notification + observable + observer
                        (notification, observableRef, observerRef)
                    )
                    disabled = False
                    if self._disabled:
                        for disableKey in holdDisabledPossibilities:
                            if disableKey in self._disabled:
                                disabled = True
                                break
                    if disabled:
                        continue
                    hold = False
                    if self._holds:
                        for holdKey in holdDisabledPossibilities:
                            if holdKey in self._holds:
                                hold = True
                                n = (notification, observableRef, data)
                                if n not in self._holds[key]["notifications"]:
                                    self._holds[holdKey]["notifications"].append(n)
                                break
                    if hold:
                        continue
                # post
                # ----
                observer = observerRef()
                if observer is None:
                    # dead ref.
                    # XXX: delete?
                    continue
                callback = getattr(observer, methodName)
                callback(notificationObj)

    # ----
    # Hold
    # ----

    def holdNotifications(self, observable=None, notification=None, observer=None, note=None):
        """
        Hold all notifications posted to all objects observing
        **notification** in **observable**.

        * **observable** The object that the notification belongs to. This is optional.
          If no *observable* is given, *all* *notifications* will be held.
        * **notification** The name of the notification. This is optional.
          If no *notification* is given, *all* notifications for *observable*
          will be held.
        * **observer** The specific observer to not hold notifications for.
          If no *observer* is given, the appropriate notifications will be
          held for all observers.
        * **note** An arbitrary string containing information about why the hold
          has been requested, the requester, etc. This is used for reference only.

        Held notifications will be posted after the matching *notification*
        and *observable* have been passed to :meth:`Notification.releaseHeldNotifications`.
        This object will retain a count of how many times it has been told to
        hold notifications for *notification* and *observable*. It will not
        post the notifications until the *notification* and *observable*
        have been released the same number of times.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        if observer is not None:
            observer = weakref.ref(observer)
        key = (notification, observable, observer)
        if key not in self._holds:
            self._holds[key] = dict(count=0, notifications=[], notes=[])
        self._holds[key]["count"] += 1
        if note is not None:
            self._holds[key]["notes"].append(note)

    def releaseHeldNotifications(self, observable=None, notification=None, observer=None):
        """
        Release all held notifications posted to all objects observing
        **notification** in **observable**.

        * **observable** The object that the notification belongs to. This is optional.
        * **notification** The name of the notification. This is optional.
        * **observer** The observer. This is optional.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        if observer is not None:
            observer = weakref.ref(observer)
        key = (notification, observable, observer)
        self._holds[key]["count"] -= 1
        if self._holds[key]["count"] == 0:
            notifications = self._holds[key]["notifications"]
            del self._holds[key]
            for notification, observableRef, data in notifications:
                self.postNotification(notification, observableRef(), data)

    def areNotificationsHeld(self, observable=None, notification=None, observer=None):
        """
        Returns a boolean indicating if notifications posted to all objects observing
        **notification** in **observable** are being held.

        * **observable** The object that the notification belongs to. This is optional.
        * **notification** The name of the notification. This is optional.
        * **observer** The observer. This is optional.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        if observer is not None:
            observer = weakref.ref(observer)
        key = (notification, observable, observer)
        return key in self._holds

    def getHeldNotifications(self):
        """
        Returns a list of all held notifications. This will be a
        tuple of the form:

        (notification, observable, observer)
        """
        return self._holds.keys()

    def getHeldNotificationNotes(self, observable=None, notification=None, observer=None):
        """
        Returns a list of notes defined for notification holds observing
        **notification** in **observable** are being held.

        * **observable** The object that the notification belongs to. This is optional.
        * **notification** The name of the notification. This is optional.
        * **observer** The observer. This is optional.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        if observer is not None:
            observer = weakref.ref(observer)
        key = (notification, observable, observer)
        return self._holds[key]["notes"]

    # -------
    # Disable
    # -------

    def disableNotifications(self, observable=None, notification=None, observer=None):
        """
        Disable all posts of **notification** from **observable** posted
        to **observer** observing.

        * **observable** The object that the notification belongs to. This is optional.
          If no *observable* is given, *all* *notifications* will be disabled for *observer*.
        * **notification** The name of the notification. This is optional.
          If no *notification* is given, *all* notifications for *observable*
          will be disabled for *observer*.
        * **observer** The specific observer to not send posts to. If no
          *observer* is given, the appropriate notifications will not
          be posted to any observers.

        This object will retain a count of how many times it has been told to
        disable notifications for *notification* and *observable*. It will not
        enable new notifications until the *notification* and *observable*
        have been released the same number of times.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        if observer is not None:
            observer = weakref.ref(observer)
        key = (notification, observable, observer)
        if key not in self._disabled:
            self._disabled[key] = 0
        self._disabled[key] += 1

    def enableNotifications(self, observable=None, notification=None, observer=None):
        """
        Enable notifications posted to all objects observing
        **notification** in **observable**.

        * **observable** The object that the notification belongs to. This is optional.
        * **notification** The name of the notification. This is optional.
        * **observer** The observer. This is optional.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        if observer is not None:
            observer = weakref.ref(observer)
        key = (notification, observable, observer)
        self._disabled[key] -= 1
        if self._disabled[key] == 0:
            del self._disabled[key]

    def areNotificationsDisabled(self, observable=None, notification=None, observer=None):
        """
        Returns a boolean indicating if notifications posted to all objects observing
        **notification** in **observable** are disabled.

        * **observable** The object that the notification belongs to. This is optional.
        * **notification** The name of the notification. This is optional.
        * **observer** The observer. This is optional.
        """
        if observable is not None:
            observable = weakref.ref(observable)
        if observer is not None:
            observer = weakref.ref(observer)
        key = (notification, observable, observer)
        return key in self._disabled

    # ----
    # Find
    # ----

    def findObservations(self, observer=None, notification=None, observable=None, identifier=None):
        """
        Find observations matching the given arguments based
        on the values that were passed during addObserver.
        A value of None for any of these indicates that all
        should be considered to match the value. In the case
        of identifier, strings will be matched using fnmatch.fnmatchcase.
        The returned value will be a list of dictionaries with
        this format:

            [
                {
                    observer=<...>
                    observable=<...>
                    methodName="..."
                    notification="..."
                    identifier="..."
                }
            ]
        """
        if observer is not None:
            observer = weakref.ref(observer)
        if observable is not None:
            observable = weakref.ref(observable)
        found = []
        for (otherNotification, otherObservable), observerDict in self._registry.items():
            if notification is not None:
                if otherNotification != notification:
                    continue
            if observable is not None:
                if otherObservable != observable:
                    continue
            for otherObserver, methodName in observerDict.items():
                otherIdentifier = None
                if (otherNotification, otherObservable) in self._identifierRegistry:
                    otherIdentifier = self._identifierRegistry[otherNotification, otherObservable].get(otherObserver)
                if otherIdentifier is None and identifier is not None:
                    continue
                if otherIdentifier is not None and identifier is not None:
                    if not fnmatchcase(otherIdentifier, identifier):
                        continue
                if observer is not None:
                    if otherObserver != observer:
                        continue
                observation = dict(
                    observer=otherObserver(),
                    observable=otherObservable(),
                    notification=otherNotification,
                    identifier=otherIdentifier
                )
                found.append(observation)
        return found


class Notification(object):

    """An object that wraps notification data."""

    __slots__ = ("_name", "_objRef", "_data")

    def __init__(self, name, objRef, data):
        self._name = name
        self._objRef = objRef
        self._data = data

    def __repr__(self):
        return "<Notification: %s %s>" % (self.name, repr(self.object))

    def _get_name(self):
        return self._name

    name = property(_get_name, doc="The notification name. A string.")

    def _get_object(self):
        if self._objRef is not None:
            return self._objRef()
        return None

    object = property(_get_object, doc="The observable object the notification belongs to.")

    def _get_data(self):
        return self._data

    data = property(_get_data, doc="Arbitrary data passed along with the notification. There is no set format for this data and there is not requirement that any data be present. Refer to the documentation for methods that are responsible for generating notifications for information about this data.")


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import unittest
from defcon.objects.color import Color


class ColorTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_from_string(self):
        self.assertEqual(tuple(Color("1,1,1,1")), (1, 1, 1, 1))
        self.assertEqual(tuple(Color(".1,.1,.1,.1")),
                         (0.10000000000000001, 0.10000000000000001,
                          0.10000000000000001, 0.10000000000000001))
        self.assertEqual(tuple(Color("1, 1, 1, 1")), (1, 1, 1, 1))
        self.assertEqual(tuple(Color("0.1, 0.1, 0.1, 0.1")),
                         (0.10000000000000001, 0.10000000000000001,
                          0.10000000000000001, 0.10000000000000001))

    def test_from_tuple(self):
        self.assertEqual(tuple(Color((1, 1, 1, 1))), (1, 1, 1, 1))

    def test_to_string(self):
        self.assertEqual(Color((0, 0, 0, 0)), "0,0,0,0")
        self.assertEqual(Color((1, .1, .01, .001)), "1,0.1,0.01,0.001")
        self.assertEqual(Color((.0001, .00001, .000001, .000005)),
                         "0.0001,0.00001,0,0.00001")

    def test_to_sequence(self):
        self.assertEqual(tuple(Color((1, 1, 1, 1))), (1, 1, 1, 1))

    def test_component_attributes(self):
        c = Color((.1, .2, .3, .4))
        self.assertEqual(c.r, 0.10000000000000001)
        self.assertEqual(c.g, 0.20000000000000001)
        self.assertEqual(c.b, 0.29999999999999999)
        self.assertEqual(c.a, 0.40000000000000002)

    def test_invalid_values_negative(self):
        with self.assertRaises(
                ValueError,
                msg="The color for r (-1) is not between 0 and 1."):
            Color((-1, 0, 0, 0))
        with self.assertRaises(
                ValueError,
                msg="The color for g (-1) is not between 0 and 1."):
            Color((0, -1, 0, 0))
        with self.assertRaises(
                ValueError,
                msg="The color for b (-1) is not between 0 and 1."):
            Color((0, 0, -1, 0))
        with self.assertRaises(
                ValueError,
                msg="The color for a (-1) is not between 0 and 1."):
            Color((0, 0, 0, -1))

    def test_invalid_values_too_large(self):
        with self.assertRaises(
                ValueError,
                msg="The color for r (2) is not between 0 and 1."):
            Color((2, 0, 0, 0))
        with self.assertRaises(
                ValueError,
                msg="The color for g (2) is not between 0 and 1."):
            Color((0, 2, 0, 0))
        with self.assertRaises(
                ValueError,
                msg="The color for b (2) is not between 0 and 1."):
            Color((0, 0, 2, 0))
        with self.assertRaises(
                ValueError,
                msg="The color for a (2) is not between 0 and 1."):
            Color((0, 0, 0, 2))

    def test_iterate_component_attributes(self):
        color = Color((0.1, 0.2, 0.3, 0.4))
        expected_values = (0.10000000000000001, 0.20000000000000001,
                           0.29999999999999999, 0.40000000000000002)
        for index, component_attribute in enumerate(color):
            self.assertEqual(component_attribute, expected_values[index])

if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
import weakref
from defcon.objects.base import BaseDictCompareObject
from defcon.objects.color import Color
from defcon.tools.identifiers import makeRandomIdentifier


class Guideline(BaseDictCompareObject):

    """
    This object represents a guideline.

    **This object posts the following notifications:**

    - Guideline.Changed
    - Guideline.XChanged
    - Guideline.YChanged
    - Guideline.AngleChanged
    - Guideline.NameChanged
    - Guideline.IdentifierChanged

    During initialization a guideline dictionary, following the format defined
    in the UFO spec, can be passed. If so, the new object will be populated
    with the data from the dictionary.
    """

    changeNotificationName = "Guideline.Changed"
    representationFactories = {}

    def __init__(self, font=None, glyph=None, guidelineDict=None):
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None
        if font is not None:
            self.font = font
        if glyph is not None:
            self.glyph = glyph
        super(Guideline, self).__init__()
        self.beginSelfNotificationObservation()
        self._dirty = False
        if guidelineDict is not None:
            self.x = guidelineDict.get("x")
            self.y = guidelineDict.get("y")
            self.angle = guidelineDict.get("angle")
            self.name = guidelineDict.get("name")
            self.color = guidelineDict.get("color")
            self.identifier = guidelineDict.get("identifier")

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        if self._glyph is not None:
            return self.glyph
        elif self._font is not None:
            return self.font
        return None

    def _get_font(self):
        font = None
        if self._font is None:
            layerSet = self.layerSet
            if layerSet is not None:
                font = layerSet.font
            if font is not None:
                self._font = weakref.ref(font)
        else:
            font = self._font()
        return font

    def _set_font(self, font):
        assert self._font is None
        assert self._glyph is None
        if font is not None:
            font = weakref.ref(font)
        self._font = font

    font = property(_get_font, _set_font, doc="The :class:`Font` that this object belongs to.")

    def _get_layerSet(self):
        layerSet = None
        if self._layerSet is None:
            layer = self.layer
            if layer is not None:
                layerSet = layer.layerSet
                if layerSet is not None:
                    self._layerSet = weakref.ref(layerSet)
        else:
            layerSet = self._layerSet()
        return layerSet

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this object belongs to (if it isn't a font info guideline).")

    def _get_layer(self):
        layer = None
        if self._layer is None:
            glyph = self.glyph
            if glyph is not None:
                layer = glyph.layer
                if layer is not None:
                    self._layer = weakref.ref(layer)
        else:
            layer = self._layer()
        return layer

    layer = property(_get_layer, doc="The :class:`Layer` that this object belongs to (if it isn't a font info guideline).")

    def _get_glyph(self):
        if self._glyph is not None:
            return self._glyph()
        return None

    def _set_glyph(self, glyph):
        assert self._font is None
        assert self._glyph is None
        if glyph is not None:
            glyph = weakref.ref(glyph)
        self._glyph = glyph

    glyph = property(_get_glyph, _set_glyph, doc="The :class:`Glyph` that this object belongs to (if it isn't a font info guideline). This should not be set externally.")

    # ----------
    # Attributes
    # ----------

    # x

    def _get_x(self):
        return self.get("x")

    def _set_x(self, value):
        old = self.get("x")
        if value == old:
            return
        self["x"] = value
        self.postNotification("Guideline.XChanged", data=dict(oldValue=old, newValue=value))

    x = property(_get_x, _set_x, doc="The x coordinate. Setting this will post *Guideline.XChanged* and *Guideline.Changed* notifications.")

    # y

    def _get_y(self):
        return self.get("y")

    def _set_y(self, value):
        old = self.get("y")
        if value == old:
            return
        self["y"] = value
        self.postNotification("Guideline.YChanged", data=dict(oldValue=old, newValue=value))

    y = property(_get_y, _set_y, doc="The y coordinate. Setting this will post *Guideline.YChanged* and *Guideline.Changed* notifications.")

    # angle

    def _get_angle(self):
        return self.get("angle")

    def _set_angle(self, value):
        old = self.get("angle")
        if value == old:
            return
        self["angle"] = value
        self.postNotification("Guideline.AngleChanged", data=dict(oldValue=old, newValue=value))

    angle = property(_get_angle, _set_angle, doc="The angle. Setting this will post *Guideline.AngleChanged* and *Guideline.Changed* notifications.")

    # name

    def _get_name(self):
        return self.get("name")

    def _set_name(self, value):
        old = self.get("name")
        if value == old:
            return
        if value is None:
            if "name" in self:
                del self["name"]
        else:
            self["name"] = value
        self.postNotification("Guideline.NameChanged", data=dict(oldValue=old, newValue=value))

    name = property(_get_name, _set_name, doc="The name. Setting this will post *Guideline.NameChanged* and *Guideline.Changed* notifications.")

    # color

    def _get_color(self):
        return self.get("color")

    def _set_color(self, color):
        if color is None:
            newColor = None
        else:
            newColor = Color(color)
        oldColor = self.get("color")
        if newColor == oldColor:
            return
        self["color"] = newColor
        self.postNotification("Guideline.ColorChanged", data=dict(oldValue=oldColor, newValue=newColor))

    color = property(_get_color, _set_color, doc="The guideline's :class:`Color` object. When setting, the value can be a UFO color string, a sequence of (r, g, b, a) or a :class:`Color` object. Setting this posts *Guideline.ColorChanged* and *Guideline.Changed* notifications.")

    # ----------
    # Identifier
    # ----------

    def _get_identifiers(self):
        identifiers = None
        parent = self.glyph
        if parent is None:
            parent = self.font
        if parent is not None:
            identifiers = parent.identifiers
        if identifiers is None:
            identifiers = set()
        return identifiers

    identifiers = property(_get_identifiers, doc="Set of identifiers for the object that this guideline belongs to. This is primarily for internal use.")

    def _get_identifier(self):
        return self.get("identifier")

    def _set_identifier(self, value):
        # don't allow overwritting an existing identifier
        if self.identifier is not None:
            return
        oldIdentifier = self.identifier
        if value == oldIdentifier:
            return
        # don't allow a duplicate
        identifiers = self.identifiers
        assert value not in identifiers
        # free the old identifier
        if oldIdentifier in identifiers:
            identifiers.remove(oldIdentifier)
        # store
        self["identifier"] = value
        if value is not None:
            identifiers.add(value)
        # post notifications
        self.postNotification("Guideline.IdentifierChanged", data=dict(oldValue=oldIdentifier, newValue=value))

    identifier = property(_get_identifier, _set_identifier, doc="The identifier. Setting this will post *Guideline.IdentifierChanged* and *Guideline.Changed* notifications.")

    def generateIdentifier(self):
        """
        Create a new, unique identifier for and assign it to the guideline.
        This will post *Guideline.IdentifierChanged* and *Guideline.Changed* notifications.
        """
        if self.identifier is None:
            identifier = makeRandomIdentifier(existing=self.identifiers)
            self.identifier = identifier
        return self.identifier

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(Guideline, self).endSelfNotificationObservation()
        self._font = None
        self._fontInfo = None
        self._layerSet = None
        self._layer = None
        self._glyph = None


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import unittest
import glob
import os
import fs
import fs.copy
import fs.path
from fontTools.ufoLib import UFOReader, UFOFileStructure
from defcon import Font, Glyph, Color, Component, Anchor, Guideline
from defcon.test.testTools import (
    getTestFontPath, makeTestFontCopy, getTestFontCopyPath,
    openTestFontAsFileSystem, closeTestFontAsFileSystem,
    tearDownTestFontCopy)

try:
    from plistlib import load, dump
except ImportError:
    from plistlib import readPlist as load, writePlist as dump


class LayerTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_set_parent_data_in_glyph(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        glyph = layer["A"]
        self.assertEqual(id(glyph.getParent()), id(font))

    def test_newGlyph(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        layer.newGlyph("NewGlyphTest")
        glyph = layer["NewGlyphTest"]
        self.assertEqual(glyph.name, "NewGlyphTest")
        self.assertTrue(glyph.dirty)
        self.assertTrue(font.dirty)
        self.assertEqual(sorted(layer.keys()), ["A", "B", "C", "NewGlyphTest"])

    def test_insertGlyph(self):
        font = Font()
        layer = font.layers[None]
        source = Glyph()
        source.unicodes = [1, 2]
        source.name = "a"
        dest = layer.insertGlyph(source, name="nota")
        self.assertNotEqual(dest, source)
        self.assertEqual(dest.name, "nota")
        self.assertEqual(list(layer.unicodeData.items()),
                         [(1, ["nota"]), (2, ["nota"])])
        source = Glyph()
        source.unicodes = [3]
        source.name = "b"
        dest = layer.insertGlyph(source)
        self.assertNotEqual(dest, source)
        self.assertEqual(dest.name, "b")
        self.assertEqual(list(layer.unicodeData.items()),
                         [(1, ["nota"]), (2, ["nota"]), (3, ["b"])])

    def test_iter(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        names = [glyph.name for glyph in layer]
        self.assertEqual(sorted(names), ["A", "B", "C"])
        names = []
        for glyph1 in layer:
            for glyph2 in layer:
                names.append((glyph1.name, glyph2.name))
        self.assertEqual(sorted(names),
                         [("A", "A"), ("A", "B"), ("A", "C"),
                          ("B", "A"), ("B", "B"), ("B", "C"),
                          ("C", "A"), ("C", "B"), ("C", "C")])

    def test_getitem(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertEqual(layer["A"].name, "A")
        self.assertEqual(layer["B"].name, "B")
        with self.assertRaises(KeyError):
            layer["NotInFont"]

    def test_len(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertEqual(len(layer), 3)

        font = Font()
        layer = font.layers["public.default"]
        self.assertEqual(len(layer), 0)

    def test_contains(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertTrue("A" in layer)
        self.assertFalse("NotInFont" in layer)

        font = Font()
        layer = font.layers["public.default"]
        self.assertFalse("A" in layer)

    def test_keys(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertEqual(sorted(layer.keys()), ["A", "B", "C"])
        del layer["A"]
        self.assertEqual(sorted(layer.keys()), ["B", "C"])
        layer.newGlyph("A")
        self.assertEqual(sorted(layer.keys()), ["A", "B", "C"])

        font = Font()
        layer = font.layers["public.default"]
        self.assertEqual(layer.keys(), set())
        layer.newGlyph("A")
        self.assertEqual(layer.keys(), {"A"})

    def test_color(self):
        font = Font(getTestFontPath())
        layer = font.layers["Layer 1"]
        self.assertIsInstance(layer.color, Color)
        self.assertEqual(str(layer.color), "0.1,0.2,0.3,0.4")
        layer.color = "0.5,0.5,0.5,0.5"
        self.assertIsInstance(layer.color, Color)
        layer.color = (.5, .5, .5, .5)
        self.assertIsInstance(layer.color, Color)
        self.assertEqual(str(layer.color), "0.5,0.5,0.5,0.5")

    def test_glyphsWithOutlines(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertEqual(sorted(layer.glyphsWithOutlines), ["A", "B"])
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        for glyph in layer:
            pass
        self.assertEqual(sorted(layer.glyphsWithOutlines), ["A", "B"])

    def test_componentReferences(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertEqual(sorted(layer.componentReferences.items()),
                         [("A", set(["C"])), ("B", set(["C"]))])
        layer["C"]
        self.assertEqual(sorted(layer.componentReferences.items()),
                         [("A", set(["C"])), ("B", set(["C"]))])

    def test_imageReferences(self):
        font = Font(getTestFontPath())
        layer = font.layers["Layer 1"]
        self.assertEqual(layer.imageReferences, {"image 1.png": ["A"]})
        layer.newGlyph("B")
        glyph = layer["B"]
        glyph.image = dict(fileName="test", xScale=1, xyScale=1,
                           yxScale=1, yScale=1, xOffset=0, yOffset=0,
                           color=None)
        self.assertEqual(sorted(layer.imageReferences.items()),
                         [("image 1.png", ["A"]), ("test", ["B"])])

    def test_bounds(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertEqual(layer.bounds, (0, 0, 700, 700))

    def test_controlPointBounds(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        self.assertEqual(layer.controlPointBounds, (0, 0, 700, 700))

    def test_lib(self):
        font = Font(getTestFontPath())
        layer = font.layers["Layer 1"]
        self.assertEqual(layer.lib, {"com.typesupply.defcon.test": "1 2 3"})
        layer.lib.dirty = False
        layer.lib["blah"] = "abc"
        self.assertEqual(layer.lib["blah"], "abc")
        self.assertTrue(layer.lib.dirty)

    def test_testForExternalChanges(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            try:
                with Font(path) as font:

                    with UFOReader(font.path) as reader:
                        self.assertEqual(font.layers[None].testForExternalChanges(reader),
                                         ([], [], []))

                        # make a simple change to a glyph
                        fileSystem = openTestFontAsFileSystem(font.path)
                        g = font.layers[None]["A"]
                        path = fs.path.join("glyphs", "A_.glif")
                        t = fileSystem.readbytes(path)
                        t += b"<!-- test -->"
                        fileSystem.writebytes(path, t)
                        g._dataOnDiskTimeStamp -= 1
                        closeTestFontAsFileSystem(fileSystem, font.path)
                    with UFOReader(font.path) as reader:
                        self.assertEqual(font.layers[None].testForExternalChanges(reader),
                                         (["A"], [], []))

                    # save the glyph and test again
                    font["A"].dirty = True
                    font.save()
                    with UFOReader(font.path) as reader:
                        self.assertEqual(font.layers[None].testForExternalChanges(reader),
                                         ([], [], []))

                    # add a glyph
                    fileSystem = openTestFontAsFileSystem(font.path)
                    path = fs.path.join("glyphs", "A_.glif")
                    t = fileSystem.readbytes(path)
                    t = t.replace(b'<glyph name="A" format="1">',
                                  b'<glyph name="XYZ" format="1">')
                    path = fs.path.join("glyphs", "XYZ.glif")
                    fileSystem.writebytes(path, t)
                    path = fs.path.join("glyphs", "contents.plist")
                    with fileSystem.open(path, "rb") as f:
                        plist = load(f)
                    savePlist = dict(plist)
                    plist["XYZ"] = "XYZ.glif"
                    with fileSystem.open(path, "wb") as f:
                        dump(plist, f)
                    closeTestFontAsFileSystem(fileSystem, font.path)
                    with UFOReader(font.path) as reader:
                        self.assertEqual(font.layers[None].testForExternalChanges(reader),
                                         ([], ["XYZ"], []))
                    path = font.path

                # delete a glyph
                with Font(path) as font:
                    g = font["XYZ"]
                    fileSystem = openTestFontAsFileSystem(font.path)
                    path = fs.path.join("glyphs", "contents.plist")
                    with fileSystem.open(path, "wb") as f:
                        dump(savePlist, f)
                    path = fs.path.join("glyphs", "XYZ.glif")
                    fileSystem.remove(path)
                    closeTestFontAsFileSystem(fileSystem, font.path)
                    with UFOReader(font.path) as reader:
                        self.assertEqual(font.layers[None].testForExternalChanges(reader),
                                         ([], [], ["XYZ"]))
            finally:
                tearDownTestFontCopy(font.path)

    def test_reloadGlyphs(self):
        path = getTestFontPath("TestExternalEditing.ufo")
        font = Font(path)
        glyph = font.layers[None]["A"]

        path = os.path.join(font.path, "glyphs", "A_.glif")
        f = open(path, "r")
        t = f.read()
        f.close()
        t = t.replace('<advance width="700"/>', '<advance width="701"/>')
        f = open(path, "w")
        f.write(t)
        f.close()

        self.assertEqual(glyph.width, 700)
        self.assertEqual(len(glyph), 2)
        font.layers[None].reloadGlyphs(["A"])
        self.assertEqual(glyph.width, 701)
        self.assertEqual(len(glyph), 2)

        t = t.replace('<advance width="701"/>', '<advance width="700"/>')
        f = open(path, "w")
        f.write(t)
        f.close()

    def test_glyph_name_change(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        glyph = layer["A"]
        glyph.name = "NameChangeTest"
        self.assertEqual(sorted(layer.keys()), ["B", "C", "NameChangeTest"])
        self.assertTrue(layer.dirty)

    def test_glyph_unicodes_changed(self):
        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        glyph = layer["A"]
        glyph.unicodes = [123, 456]
        self.assertEqual(layer.unicodeData[123], ["A"])
        self.assertEqual(layer.unicodeData[456], ["A"])
        self.assertEqual(layer.unicodeData[66], ["B"])
        self.assertIsNone(layer.unicodeData.get(65))

        font = Font(getTestFontPath())
        layer = font.layers["public.default"]
        layer.newGlyph("test")
        glyph = layer["test"]
        glyph.unicodes = [65]
        self.assertEqual(layer.unicodeData[65], ["test", "A"])

    def test_glyph_dispatcher_loaded(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertIsNotNone(glyph.dispatcher)
        self.assertEqual(glyph.dispatcher, font.dispatcher)
        contour = glyph[0]
        self.assertEqual(contour.getParent(), glyph)
        self.assertEqual(contour.dispatcher, font.dispatcher)
        anchor = glyph.anchors[0]
        self.assertEqual(anchor.getParent(), glyph)
        self.assertEqual(anchor.dispatcher, font.dispatcher)
        glyph = font["C"]
        component = glyph.components[0]
        self.assertEqual(component.getParent(), glyph)
        self.assertEqual(component.dispatcher, font.dispatcher)
        glyph = font.layers["Layer 1"]["A"]
        guideline = glyph.guidelines[0]
        self.assertEqual(guideline.getParent(), glyph)
        self.assertEqual(guideline.dispatcher, font.dispatcher)

    def test_glyph_dispatcher_new(self):
        font = Font()
        font.newGlyph("A")
        glyph = font["A"]
        pen = glyph.getPointPen()
        pen.beginPath()
        pen.addPoint((0, 0), segmentType="line")
        pen.addPoint((0, 100), segmentType="line")
        pen.addPoint((100, 100), segmentType="line")
        pen.addPoint((100, 0), segmentType="line")
        pen.endPath()
        contour = glyph[0]
        self.assertEqual(contour.getParent(), glyph)
        self.assertEqual(contour.dispatcher, font.dispatcher)
        component = Component()
        glyph.appendComponent(component)
        self.assertEqual(component.getParent(), glyph)
        self.assertEqual(component.dispatcher, font.dispatcher)
        anchor = Anchor()
        glyph.appendAnchor(anchor)
        self.assertEqual(anchor.getParent(), glyph)
        self.assertEqual(anchor.dispatcher, font.dispatcher)
        guideline = Guideline()
        glyph.appendGuideline(guideline)
        self.assertEqual(guideline.getParent(), glyph)
        self.assertEqual(guideline.dispatcher, font.dispatcher)

    def test_glyph_dispatcher_inserted(self):
        font = Font()
        font.newGlyph("A")
        glyph = font["A"]
        pen = glyph.getPointPen()
        pen.beginPath()
        pen.addPoint((0, 0), segmentType="line")
        pen.addPoint((0, 100), segmentType="line")
        pen.addPoint((100, 100), segmentType="line")
        pen.addPoint((100, 0), segmentType="line")
        pen.endPath()
        contour = glyph[0]
        component = Component()
        glyph.appendComponent(component)
        anchor = Anchor()
        glyph.appendAnchor(anchor)
        guideline = Guideline()
        glyph.appendGuideline(guideline)
        sourceGlyph = glyph
        newFont = Font()
        insertedGlyph = newFont.insertGlyph(sourceGlyph)
        contour = insertedGlyph[0]
        self.assertTrue(contour.getParent(), insertedGlyph)
        self.assertTrue(contour.dispatcher, newFont.dispatcher)
        component = insertedGlyph.components[0]
        self.assertTrue(component.getParent(), insertedGlyph)
        self.assertTrue(component.dispatcher, newFont.dispatcher)
        anchor = insertedGlyph.anchors[0]
        self.assertTrue(anchor.getParent(), insertedGlyph)
        self.assertTrue(anchor.dispatcher, newFont.dispatcher)
        guideline = insertedGlyph.guidelines[0]
        self.assertTrue(guideline.getParent(), insertedGlyph)
        self.assertTrue(guideline.dispatcher, newFont.dispatcher)

    def test_tempLib(self):
        font = Font()
        layer = font.layers["public.default"]
        layer.tempLib["foo"] = "bar"
        self.assertEqual(layer.tempLib, {"foo": "bar"})

        otherLayer = font.newLayer("other.layer")
        otherLayer.setDataFromSerialization(layer.getDataForSerialization())
        self.assertEqual(otherLayer.tempLib, {"foo": "bar"})


class LayerWithTestFontCopyTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.path = makeTestFontCopy()

    def tearDown(self):
        if os.path.exists(getTestFontCopyPath()):
            tearDownTestFontCopy()

    def test_delitem(self):
        path = self.path
        font = Font(path)
        layer = font.layers["public.default"]
        # glyph = layer["A"]
        self.assertTrue("A" in layer)
        del layer["A"]
        self.assertFalse("A" in layer)
        self.assertTrue(layer.dirty)
        layer.newGlyph("NewGlyphTest")
        del layer["NewGlyphTest"]
        self.assertEqual(sorted(layer.keys()), ["B", "C"])
        self.assertEqual(len(layer), 2)
        self.assertFalse("A" in layer)
        font.save()
        fileNames = glob.glob(os.path.join(path, "glyphs", "*.glif"))
        fileNames = [os.path.basename(fileName) for fileName in fileNames]
        self.assertEqual(sorted(fileNames), ["B_.glif", "C_.glif"])
        with self.assertRaises(KeyError):
            del layer["NotInFont"]

    # def test_delitem_glyph_not_dirty(self):
    #     path = self.path
    #     font = Font(path)
    #     layer = font.layers["public.default"]
    #     # glyph = layer["A"]
    #     glyphPath = os.path.join(path, "glyphs", "A_.glif")
    #     os.remove(glyphPath)
    #     contentsPath = os.path.join(path, "glyphs", "contents.plist")
    #     with open(contentsPath, "rb") as f:
    #         plist = load(f)
    #     del plist["A"]
    #     with open(contentsPath, "wb") as f:
    #         dump(plist, f)
    #     r = font.testForExternalChanges()
    #     self.assertEqual(r["deletedGlyphs"], ["A"])
    #     del layer["A"]
    #     font.save()
    #     self.assertFalse(os.path.exists(glyphPath))

    # def test_delitem_glyph_dirty(self):
    #     path = self.path
    #     font = Font(path)
    #     layer = font.layers["public.default"]
    #     glyph = layer["A"]
    #     glyph.dirty = True
    #     glyphPath = os.path.join(path, "glyphs", "A_.glif")
    #     os.remove(glyphPath)
    #     contentsPath = os.path.join(path, "glyphs", "contents.plist")
    #     with open(contentsPath, "rb") as f:
    #         plist = load(f)
    #     del plist["A"]
    #     with open(contentsPath, "wb") as f:
    #         dump(plist, f)
    #     r = font.testForExternalChanges()
    #     self.assertEqual(r["deletedGlyphs"], ["A"])
    #     del layer["A"]
    #     font.save()
    #     self.assertFalse(os.path.exists(glyphPath))


if __name__ == "__main__":
    unittest.main()



import unittest
from defcon.test.testTools import getTestFontPath
from defcon.objects.font import Font


class KerningTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_keys(self):
        font = Font(getTestFontPath())
        self.assertEqual(sorted(font.kerning.keys()),
                         [("A", "A"), ("A", "B")])

    def test_items(self):
        font = Font(getTestFontPath())
        self.assertEqual(sorted(font.kerning.items()),
                         [(("A", "A"), -100), (("A", "B"), 100)])

    def test_values(self):
        font = Font(getTestFontPath())
        self.assertEqual(sorted(font.kerning.values()),
                         [-100, 100])

    def test___contains__(self):
        font = Font(getTestFontPath())
        self.assertIn(("A", "B"), font.kerning)
        self.assertNotIn(("NotInFont", "NotInFont"), font.kerning)

    def test_get(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.kerning.get(("A", "A")), -100)
        self.assertEqual(font.kerning.get(("NotInFont", "NotInFont"), 0), 0)

    def test___getitem__(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.kerning["A", "A"], -100)
        with self.assertRaisesRegex(KeyError, "\('NotInFont', 'NotInFont'\)"):
            font.kerning["NotInFont", "NotInFont"]

    def test___setitem__(self):
        font = Font(getTestFontPath())
        font.kerning["NotInFont", "NotInFont"] = 100
        self.assertEqual(sorted(font.kerning.keys()),
                         [("A", "A"), ("A", "B"), ("NotInFont", "NotInFont")])
        self.assertTrue(font.kerning.dirty)

    def test_clear(self):
        font = Font(getTestFontPath())
        font.kerning.clear()
        self.assertEqual(list(font.kerning.keys()), [])
        self.assertTrue(font.kerning.dirty)

    def test_update(self):
        font = Font(getTestFontPath())
        other = {("X", "X"): 500}
        font.kerning.update(other)
        self.assertEqual(sorted(font.kerning.keys()),
                         [("A", "A"), ("A", "B"), ("X", "X")])
        self.assertTrue(font.kerning.dirty)

    def test_find(self):
        font = Font()
        font.groups["public.kern1.A"] = ["A", "A.alt"]
        font.groups["public.kern2.C"] = ["C", "C.alt"]
        font.kerning["public.kern1.A", "public.kern2.C"] = 1
        font.kerning["public.kern1.A", "C.alt"] = 2
        font.kerning["A.alt", "C.alt"] = 3
        self.assertEqual(font.kerning.find(("A", "C")), 1)
        self.assertEqual(font.kerning.find(("A", "C.alt")), 2)
        self.assertEqual(font.kerning.find(("A.alt", "C.alt")), 3)

if __name__ == "__main__":
    unittest.main()




class Color(str):

    """
    This object represents a color. This object is immutable.

    The initial argument can be either a color string as defined in the UFO
    specification or a sequence of (red, green, blue, alpha) components.

    By calling str(colorObject) you will get a UFO compatible color string.
    You can also iterate over the object to create a sequence::

        colorTuple = tuple(colorObject)
    """

    def __new__(self, value):
        # convert from string
        if isinstance(value, str):
            value = _stringToSequence(value)
        r, g, b, a = value
        # validate the values
        color = (("r", r), ("g", g), ("b", b), ("a", a))
        for component, v in color:
            if v < 0 or v > 1:
                raise ValueError("The color for %s (%s) is not between 0 and 1." % (component, str(v)))
        # convert back to a normalized string
        r = _stringify(r)
        g = _stringify(g)
        b = _stringify(b)
        a = _stringify(a)
        s = ",".join((r, g, b, a))
        # call the super
        return super(Color, self).__new__(Color, s)

    def __iter__(self):
        value = _stringToSequence(self)
        return iter(value)

    def _get_r(self):
        return _stringToSequence(self)[0]

    r = property(_get_r, "The red component.")

    def _get_g(self):
        return _stringToSequence(self)[1]

    g = property(_get_g, "The green component.")

    def _get_b(self):
        return _stringToSequence(self)[2]

    b = property(_get_b, "The blue component.")

    def _get_a(self):
        return _stringToSequence(self)[3]

    a = property(_get_a, "The alpha component.")


def _stringToSequence(value):
    r, g, b, a = [i.strip() for i in value.split(",")]
    value = []
    for component in (r, g, b, a):
        try:
            v = int(component)
            value.append(v)
            continue
        except ValueError:
            pass
        v = float(component)
        value.append(v)
    return value


def _stringify(v):
    """
    >>> _stringify(1)
    '1'
    >>> _stringify(.1)
    '0.1'
    >>> _stringify(.01)
    '0.01'
    >>> _stringify(.001)
    '0.001'
    >>> _stringify(.0001)
    '0.0001'
    >>> _stringify(.00001)
    '0.00001'
    >>> _stringify(.000001)
    '0'
    >>> _stringify(.000005)
    '0.00001'
    """
    # it's an int
    i = int(v)
    if v == i:
        return str(i)
    # it's a float
    else:
        # find the shortest possible float
        for i in range(1, 6):
            s = "%%.%df" % i
            s = s % v
            if float(s) == v:
                break
        # see if the result can be converted to an int
        f = float(s)
        i = int(f)
        if f == i:
            return str(i)
        # otherwise return the float
        return s


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import unittest
from defcon import Font, Info, Kerning, Groups
from defcon.objects.base import setUfoLibReadValidate, setUfoLibWriteValidate
from defcon.test.testTools import getTestFontPath


class UFOReadWriteValidateTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_ufoLibReadValidate_defaults(self):
        font = Font()
        self.assertTrue(font.ufoLibReadValidate)
        self.assertTrue(font.ufoLibWriteValidate)
        self.assertTrue(font.info.ufoLibReadValidate)
        self.assertTrue(font.info.ufoLibWriteValidate)

    def test_ufoLibReadValidate_settingDefaults(self):
        setUfoLibReadValidate(False)
        setUfoLibWriteValidate(False)
        font = Font()
        self.assertFalse(font.ufoLibReadValidate)
        self.assertFalse(font.ufoLibWriteValidate)
        self.assertFalse(font.info.ufoLibReadValidate)
        self.assertFalse(font.info.ufoLibWriteValidate)

        setUfoLibReadValidate(False)
        setUfoLibWriteValidate(True)
        font = Font()
        self.assertFalse(font.ufoLibReadValidate)
        self.assertTrue(font.ufoLibWriteValidate)
        self.assertFalse(font.info.ufoLibReadValidate)
        self.assertTrue(font.info.ufoLibWriteValidate)

        setUfoLibReadValidate(True)
        setUfoLibWriteValidate(False)
        font = Font()
        self.assertTrue(font.ufoLibReadValidate)
        self.assertFalse(font.ufoLibWriteValidate)
        self.assertTrue(font.info.ufoLibReadValidate)
        self.assertFalse(font.info.ufoLibWriteValidate)

    def test_customClasses(self):

        class CustomInfo(Info):

            ufoLibReadValidate = False
            ufoLibWriteValidate = False

        class CustomKerning(Kerning):

            ufoLibReadValidate = True
            ufoLibWriteValidate = False

        font = Font(infoClass=CustomInfo, kerningClass=CustomKerning)
        self.assertTrue(font.ufoLibReadValidate)
        self.assertTrue(font.ufoLibWriteValidate)
        self.assertFalse(font.info.ufoLibReadValidate)
        self.assertFalse(font.info.ufoLibWriteValidate)
        self.assertTrue(font.kerning.ufoLibReadValidate)
        self.assertFalse(font.kerning.ufoLibWriteValidate)


if __name__ == "__main__":
    unittest.main()



from __future__ import unicode_literals
import unittest
from defcon import Component, Font, Glyph
from defcon.test.testTools import NotificationTestObserver
from defcon.test.testTools import getTestFontPath


class ComponentTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.font = Font()
        self.glyph = Glyph()
        self.component = Component(self.glyph)

    def tearDown(self):
        del self.component
        del self.glyph

    def test_getParent(self):
        self.assertEqual(self.component.getParent(), self.glyph)

    def test_font(self):
        self.assertIsNone(self.component.font)
        self.component = Component(self.font.newGlyph("A"))
        self.assertEqual(self.component.font, self.font)
        with self.assertRaises(AttributeError):
            self.component.font = "foo"

    def test_layerSet(self):
        self.assertIsNone(self.component.layerSet)
        self.glyph = self.font.newGlyph("A")
        self.component = Component(self.glyph)
        self.assertEqual(self.component.layerSet, self.glyph.layerSet)
        with self.assertRaises(AttributeError):
            self.component.layerSet = "foo"

    def test_layer(self):
        self.assertIsNone(self.component.layer)
        self.glyph = self.font.newGlyph("A")
        self.component = Component(self.glyph)
        self.assertEqual(self.component.layer, self.glyph.layer)
        with self.assertRaises(AttributeError):
            self.component.layer = "foo"

    def test_glyph(self):
        self.assertIsNone(self.component.layer)
        self.glyph = self.font.newGlyph("A")
        self.component = Component(self.glyph)
        self.assertEqual(self.component.layer, self.glyph.layer)
        with self.assertRaises(AttributeError):
            self.component.layer = "foo"

    def test_bounds(self):
        self.font = Font(getTestFontPath())
        self.glyph = self.font["C"]
        self.component = self.glyph.components[0]
        self.assertEqual(
            self.component.bounds,
            (0.0, 0.0, 350.0, 350.0)
        )
        with self.assertRaises(AttributeError):
            self.component.layer = (0.0, 0.0, 350.0, 350.0)

    def test_controlPointBounds(self):
        self.font = Font(getTestFontPath())
        self.glyph = self.font["C"]
        self.component = self.glyph.components[0]
        self.assertEqual(
            self.component.controlPointBounds,
            (0.0, 0.0, 350.0, 350.0)
        )
        with self.assertRaises(AttributeError):
            self.component.layer = (0.0, 0.0, 350.0, 350.0)

    def test_baseGlyph(self):
        self.assertIsNone(self.component.baseGlyph)

        self.font = Font(getTestFontPath())
        self.glyph = self.font["C"]
        self.component = self.glyph.components[0]
        self.assertEqual(self.component.baseGlyph, "A")
        self.component.baseGlyph = "B"
        self.assertEqual(self.component.baseGlyph, "B")

    def test_transformation(self):
        self.assertEqual(
            self.component.transformation,
            (1, 0, 0, 1, 0, 0)
        )

        self.font = Font(getTestFontPath())
        self.glyph = self.font["C"]
        self.component = self.glyph.components[0]
        self.assertEqual(
            self.component.transformation,
            (0.5, 0, 0, 0.5, 0, 0)
        )
        self.component = self.glyph.components[1]
        self.assertEqual(
            self.component.transformation,
            (0.5, 0, 0, 0.5, 350, 350)
        )

    def test_identifier(self):
        self.assertIsNone(self.component.identifier)
        self.component.identifier = "component 1"
        self.assertEqual(self.component.identifier, "component 1")

    def test_identifiers(self):
        self.assertEqual(sorted(self.glyph.identifiers), [])
        self.component.identifier = "component 1"
        self.assertEqual(sorted(self.glyph.identifiers), ["component 1"])

    def test_duplicate_identifier_error(self):
        glyph = self.glyph
        component = self.component
        component.identifier = "component 1"
        self.assertEqual(component.identifier, "component 1")
        component = Component(glyph)
        with self.assertRaises(AssertionError):
            component.identifier = "component 1"
        component.identifier = "component 2"
        self.assertEqual(sorted(glyph.identifiers),
                         ["component 1", "component 2"])
        component.identifier = "not component 2 anymore"
        self.assertEqual(component.identifier, "component 2")
        self.assertEqual(sorted(glyph.identifiers),
                         ["component 1", "component 2"])
        component.identifier = None
        self.assertEqual(component.identifier, "component 2")
        self.assertEqual(sorted(glyph.identifiers),
                         ["component 1", "component 2"])


class ComponentNotificationTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.font = Font()
        self.glyph = self.font.newGlyph("A")
        # self.font.newGlyph("B")
        self.component = Component(self.glyph)
        self.component.name = "component1"
        self.notificationObject = NotificationTestObserver()

    def tearDown(self):
        del self.component
        del self.glyph
        del self.font
        del self.notificationObject

    def test_changed_notification(self):
        self.component.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Component.Changed",
            observable=self.component
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.component.baseGlyph = "B"
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Component.Changed", "component1")
        )

    # TODO: other attributes

    def test_baseGlyph_changed_notification(self):
        self.component.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Component.BaseGlyphChanged",
            observable=self.component
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.component.baseGlyph = "B"
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Component.BaseGlyphChanged", "component1")
        )

    def test_transformation_changed_notification(self):
        self.component.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Component.TransformationChanged",
            observable=self.component
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.component.transformation = (1, 0, 0, 1.1, 0, 0)
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Component.TransformationChanged", "component1")
        )

    def test_identifier_changed_notification(self):
        self.component.dispatcher.addObserver(
            observer=self.notificationObject,
            methodName="notificationCallback",
            notification="Component.IdentifierChanged",
            observable=self.component
        )
        self.assertEqual(self.notificationObject.stack, [])
        self.component.identifier = "component1_identifier"
        self.assertEqual(
            self.notificationObject.stack[-1],
            ("Component.IdentifierChanged", "component1")
        )

if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
import weakref
from warnings import warn
from fontTools.misc import bezierTools
from fontTools.misc import arrayTools
from defcon.objects.base import BaseObject
from defcon.tools import bezierMath
from defcon.tools.representations import contourBoundsRepresentationFactory,\
    contourControlPointBoundsRepresentationFactory, contourAreaRepresentationFactory,\
    contourFlattenedRepresentationFactory
from defcon.tools.identifiers import makeRandomIdentifier


class Contour(BaseObject):

    """
    This object represents a contour and it contains a list of points.

    **This object posts the following notifications:**

    - Contour.Changed
    - Contour.WindingDirectionChanged
    - Contour.PointsChanged
    - Contour.IdentifierChanged

    The Contour object has list like behavior. This behavior allows you to interact
    with point data directly. For example, to get a particular point::

        point = contour[0]

    To iterate over all points::

        for point in contour:

    To get the number of points::

        pointCount = len(contour)

    To interact with components or anchors in a similar way,
    use the ``components`` and ``anchors`` attributes.
    """

    changeNotificationName = "Contour.Changed"
    representationFactories = {
        "defcon.contour.bounds" : dict(
            factory=contourBoundsRepresentationFactory,
            destructiveNotifications=("Contour.PointsChanged")
        ),
        "defcon.contour.controlPointBounds" : dict(
            factory=contourControlPointBoundsRepresentationFactory,
            destructiveNotifications=("Contour.PointsChanged")
        ),
        "defcon.contour.area" : dict(
            factory=contourAreaRepresentationFactory,
            destructiveNotifications=("Contour.PointsChanged", "Contour.WindingDirectionChanged")
        ),
        "defcon.contour.flattened" : dict(
            factory=contourFlattenedRepresentationFactory,
            destructiveNotifications=("Contour.PointsChanged", "Contour.WindingDirectionChanged")
        )
    }

    def __init__(self, glyph=None, pointClass=None):
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None
        self.glyph = glyph
        super(Contour, self).__init__()
        self.beginSelfNotificationObservation()
        self._points = []
        if pointClass is None:
            from .point import Point
            pointClass = Point
        self._pointClass = pointClass
        self._identifier = None
        self._dirty = False

    def __del__(self):
        super(Contour, self).__del__()
        self._points = None

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.glyph

    def _get_font(self):
        font = None
        if self._font is None:
            glyph = self.glyph
            if glyph is not None:
                font = glyph.font
                if font is not None:
                    self._font = weakref.ref(font)
        else:
            font = self._font()
        return font

    font = property(_get_font, doc="The :class:`Font` that this contour belongs to.")

    def _get_layerSet(self):
        layerSet = None
        if self._layerSet is None:
            glyph = self.glyph
            if glyph is not None:
                layerSet = glyph.layerSet
                if layerSet is not None:
                    self._layerSet = weakref.ref(layerSet)
        else:
            layerSet = self._layerSet()
        return layerSet

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this contour belongs to.")

    def _get_layer(self):
        layer = None
        if self._layer is None:
            glyph = self.glyph
            if glyph is not None:
                layer = glyph.layer
                if layer is not None:
                    self._layer = weakref.ref(layer)
        else:
            layer = self._layer()
        return layer

    layer = property(_get_layer, doc="The :class:`Layer` that this contour belongs to.")

    def _get_glyph(self):
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph):
        assert self._glyph is None
        if glyph is not None:
            glyph = weakref.ref(glyph)
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = glyph

    glyph = property(_get_glyph, _set_glyph, doc="The :class:`Glyph` that this contour belongs to. This should not be set externally.")

    # ------
    # Points
    # ------

    def _get_pointClass(self):
        return self._pointClass

    pointClass = property(_get_pointClass, doc="The class used for point.")

    def _get_onCurvePoints(self):
        return [point for point in self._points if point.segmentType]

    onCurvePoints = property(_get_onCurvePoints, doc="A list of all on curve points in the contour.")

    def appendPoint(self, point):
        """
        Append **point** to the glyph. The point must be a defcon
        :class:`Point` object or a subclass of that object. An error
        will be raised if the point's identifier conflicts with any of
        the identifiers within the glyph.

        This will post *Contour.PointsChanged* and *Contour.Changed* notifications.
        """
        assert point not in self._points
        self.insertPoint(len(self._points), point)

    def insertPoint(self, index, point):
        """
        Insert **point** into the contour at index. The point
        must be a defcon :class:`Point` object or a subclass
        of that object. An error will be raised if the points's
        identifier conflicts with any of the identifiers within
        the glyph.

        This will post *Contour.PointsChanged* and *Contour.Changed* notifications.
        """
        assert point not in self._points
        if point.identifier is not None:
            identifiers = self.identifiers
            assert point.identifier not in identifiers
            if point.identifier is not None:
                identifiers.add(point.identifier)
        self._points.insert(index, point)
        self.postNotification("Contour.PointsChanged")
        self.dirty = True

    def removePoint(self, point):
        """
        Remove **point** from the contour.

        This will post *Contour.PointsChanged* and *Contour.Changed* notifications.
        """
        self._points.remove(point)
        if point.identifier is not None:
            self.identifiers.remove(point.identifier)
        self.postNotification("Contour.PointsChanged")
        self.dirty = True

    def setStartPoint(self, index):
        """
        Set the point at **index** as the first point in the contour.
        This point must be an on-curve point.

        This will post *Contour.PointsChanged* and *Contour.Changed* notifications.
        """
        onCurvePoints = self.onCurvePoints
        if len(onCurvePoints) < 2:
            return
        if self.open:
            return
        point = self._points[index]
        assert point.segmentType is not None, "index must represent an on curve point"
        before = self._points[:index]
        self._points = self._points[index:] + before
        self.postNotification("Contour.PointsChanged")
        self.dirty = True

    # -------------
    # List Behavior
    # -------------

    def __len__(self):
        return len(self._points)

    def __getitem__(self, index):
        return self._points.__getitem__(index)

    def __iter__(self):
        return iter(self._points)

    def clear(self):
        """
        Clear the contents of the contour.

        This posts *Contour.PointsChanged* and *Contour.Changed* notifications.
        """
        self._clear()

    def _clear(self, postNotification=True):
        # clear the internal storage
        self._points = []
        # reset the clockwise cache
        # post a dirty notification
        if postNotification:
            self.postNotification("Contour.PointsChanged")
            self.dirty = True

    def index(self, point):
        """
        Get the index for **point**.
        """
        return self._points.index(point)

    def reverse(self):
        """
        Reverse the direction of the contour. It's important to note
        that the actual points stored in this object will be completely
        replaced by new points.

        This will post *Contour.WindingDirectionChanged*,
        *Contour.PointsChanged* and *Contour.Changed* notifications.
        """
        from fontTools.pens.pointPen import ReverseContourPointPen
        oldDirection = self.clockwise
        # put the current points in another contour
        otherContour = self.__class__(glyph=None, pointClass=self.pointClass)
        # draw the points in this contour through
        # the reversing pen.
        reversePen = ReverseContourPointPen(otherContour)
        self.drawPoints(reversePen)
        # clear the points in this contour
        self._clear(postNotification=False)
        # set the points back into this contour
        self._points = otherContour._points
        # post a notification
        self.postNotification("Contour.WindingDirectionChanged", data=dict(oldValue=oldDirection, newValue=self.clockwise))
        self.postNotification("Contour.PointsChanged")
        self.dirty = True

    # --------
    # Segments
    # --------

    def _get_segments(self):
        if not len(self._points):
            return []
        segments = [[]]
        lastWasOffCurve = False
        for point in self._points:
            segments[-1].append(point)
            if point.segmentType is not None:
                segments.append([])
            lastWasOffCurve = point.segmentType is None
        if len(segments[-1]) == 0:
            del segments[-1]
        if lastWasOffCurve:
            if len(segments) != 1:
                lastSegment = segments[-1]
                segment = segments.pop(0)
                lastSegment.extend(segment)
        elif segments[0][-1].segmentType != "move":
            segment = segments.pop(0)
            segments.append(segment)
        return segments

    segments = property(_get_segments, doc="A list of all points in the contour organized into segments.")

    def removeSegment(self, segmentIndex, preserveCurve=False):
        """
        Remove the segment at **segmentIndex**. If
        **preserveCurve** is True, the contour will
        try to preserve the overall curve shape.
        """
        segments = self.segments
        nextIndex = segmentIndex + 1
        if nextIndex == len(segments):
            nextIndex = 0
        previousIndex = segmentIndex - 1
        if previousIndex < 0:
            previousIndex = len(segments) + previousIndex
        nextSegment = segments[nextIndex]
        segment = segments[segmentIndex]
        previousSegment = segments[previousIndex]
        # if preserveCurve is off
        # or if all are lines, handle it
        if not preserveCurve or (previousSegment[-1].segmentType == "line"\
            and segment[-1].segmentType == "line"\
            and nextSegment[-1].segmentType == "line"):
            for point in segment:
                self._points.remove(point)
            # if we're removing a move segment, we need to forward the move to
            # the next on curve
            if segment[-1].segmentType == "move":
                for point in nextSegment[:-1]:
                    self._points.remove(point)
                nextSegment[-1].segmentType = "move"
        # if have a curve, do the preservation
        else:
            # gather the needed points
            previousOnCurveX = previousSegment[-1].x
            previousOnCurveY = previousSegment[-1].y
            onCurveX = segment[-1].x
            onCurveY = segment[-1].y
            nextOnCurveX = nextSegment[-1].x
            nextOnCurveY = nextSegment[-1].y
            if segment[-1].segmentType == "curve":
                offCurve1X = segment[0].x
                offCurve1Y = segment[0].y
                offCurve2X = segment[-2].x
                offCurve2Y = segment[-2].y
            elif segment[-1].segmentType == "line":
                offCurve1X = previousOnCurveX
                offCurve1Y = previousOnCurveY
                offCurve2X = onCurveX
                offCurve2Y = onCurveY
            else:
                # XXX could be a quad. in that case, we can't handle it.
                raise NotImplementedError("unknown segment type: %s" % segment[-1].segmentType)
            if nextSegment[-1].segmentType == "curve":
                nextOffCurve1X = nextSegment[0].x
                nextOffCurve1Y = nextSegment[0].y
                nextOffCurve2X = nextSegment[-2].x
                nextOffCurve2Y = nextSegment[-2].y
            elif nextSegment[-1].segmentType == "line":
                nextOffCurve1X = onCurveX
                nextOffCurve1Y = onCurveY
                nextOffCurve2X = nextOnCurveX
                nextOffCurve2Y = nextOnCurveY
            else:
                # XXX could be a quad. in that case, we can't handle it.
                raise NotImplementedError("unknown segment type: %s" % nextSegment[-1].segmentType)
            # now do the math
            result = bezierMath.joinSegments((previousOnCurveX, previousOnCurveY),
                (offCurve1X, offCurve1Y), (offCurve2X, offCurve2Y), (onCurveX, onCurveY),
                (nextOffCurve1X, nextOffCurve1Y), (nextOffCurve2X, nextOffCurve2Y), (nextOnCurveX, nextOnCurveY))
            # remove the segment
            for point in segment:
                self._points.remove(point)
            # if the next segment type isn't a curve, make it one
            if not nextSegment[-1].segmentType == "curve":
                nextSegment[-1].segmentType = "curve"
                pointIndex = self._points.index(nextSegment[-1])
                newPoints = [self._pointClass((result[0][0], result[0][1])), self._pointClass((result[1][0], result[1][1]))]
                if pointIndex == 0:
                    self._points.extend(newPoints)
                else:
                    self._points = self._points[:pointIndex] + newPoints + self._points[pointIndex:]
            # otherwise, set the point positions
            else:
                nextSegment[0].x = result[0][0]
                nextSegment[0].y = result[0][1]
                nextSegment[1].x = result[1][0]
                nextSegment[1].y = result[1][1]
        # mark the contour as dirty
        self.dirty = True

    # ----------------
    # Basic Attributes
    # ----------------

    # clockwise

    def _get_clockwise(self):
        area = self.getRepresentation("defcon.contour.area")
        return area < 0

    def _set_clockwise(self, value):
        if self.clockwise != value:
            self.reverse()
            self._clockwiseCache = None

    clockwise = property(_get_clockwise, _set_clockwise, doc="A boolean representing if the contour has a clockwise direction. Setting this posts *Contour.WindingDirectionChanged* and *Contour.Changed* notifications.")

    # open

    def _get_open(self):
        if not self._points:
            return True
        return self._points[0].segmentType == 'move'

    open = property(_get_open, doc="A boolean indicating if the contour is open or not.")

    # ------
    # Bounds
    # ------

    def _get_bounds(self):
        return self.getRepresentation("defcon.contour.bounds")

    bounds = property(_get_bounds, doc="The bounds of the contour's outline expressed as a tuple of form (xMin, yMin, xMax, yMax).")

    def _get_controlPointBounds(self):
        return self.getRepresentation("defcon.contour.controlPointBounds")

    controlPointBounds = property(_get_controlPointBounds, doc="The control bounds of all points in the contour. This only measures the point positions, it does not measure curves. So, curves without points at the extrema will not be properly measured.")

    # ----
    # Area
    # ----

    def _get_area(self):
        return abs(self.getRepresentation("defcon.contour.area"))

    area = property(_get_area, doc="The area of the contour's outline.")

    # ----
    # Move
    # ----

    def move(self, values):
        """
        Move all points in the contour by **(x, y)**.

        This will post *Contour.PointsChanged* and *Contour.Changed* notifications.
        """
        (x, y) = values
        for point in self._points:
            point.move((x, y))
        # update the representations
        # XXX this is strictly against the rules.
        # XXX subclasses should never, ever do
        # XXX anything like this. this is a *very*
        # XXX special case.
        if "defcon.contour.bounds" in self._representations:
            bounds = self._representations["defcon.contour.bounds"][None]
            if bounds is not None:
                xMin, yMin, xMax, yMax = bounds
                xMin += x
                yMin += y
                xMax += x
                yMax += y
                bounds = (xMin, yMin, xMax, yMax)
            self._representations["defcon.contour.bounds"][None] = bounds
        if "defcon.contour.controlPointBounds" in self._representations:
            bounds = self._representations["defcon.contour.controlPointBounds"][None]
            if bounds is not None:
                xMin, yMin, xMax, yMax = bounds
                xMin += x
                yMin += y
                xMax += x
                yMax += y
                bounds = (xMin, yMin, xMax, yMax)
            self._representations["defcon.contour.controlPointBounds"][None] = bounds
        self.disableNotifications(observer=self)
        self.postNotification("Contour.PointsChanged")
        self.enableNotifications(observer=self)
        self.dirty = True

    # ------------
    # Point Inside
    # ------------

    def pointInside(self, coordinates, evenOdd=False):
        """
        Returns a boolean indicating if **(x, y)** is in the
        "black" area of the contour.
        """
        (x, y) = coordinates
        from fontTools.pens.pointInsidePen import PointInsidePen
        pen = PointInsidePen(glyphSet=None, testPoint=(x, y), evenOdd=evenOdd)
        self.draw(pen)
        return pen.getResult()

    # --------------
    # Contour Inside
    # --------------

    def contourInside(self, other, segmentLength=10):
        """
        Returns a boolean indicating if **other** is in the
        "black" area of the contour. This uses a flattened
        version of other's curves to calculate the location
        of the curves within this contour. **segmentLength**
        defines the desired length for the flattening process.
        A lower value will yeild higher accuracy but will require
        more computation time.
        """
        if segmentLength < 1:
            segmentLength = 1
        # test bounding boxes for intersection
        rect1 = self.bounds
        rect2 = other.bounds
        if not arrayTools.sectRect(rect1, rect2)[0]:
            return False
        # test existing on curves
        testedPoints = set()
        for point in other:
            if point.segmentType is None:
                continue
            pt = (point.x, point.y)
            if pt in testedPoints:
                continue
            if not self.pointInside(pt):
                return False
            testedPoints.add(pt)
        # flatten into line and test new points
        flat2 = other.getRepresentation("defcon.contour.flattened", approximateSegmentLength=segmentLength, segmentLines=True)
        for point in flat2:
            pt = (point.x, point.y)
            if pt in testedPoints:
                continue
            if not self.pointInside(pt):
                return False
            testedPoints.add(pt)
        return True

    # ---------
    # Splitting
    # ---------

    def positionForProspectivePointInsertionAtSegmentAndT(self, segmentIndex, t):
        """
        Get the precise coordinates and a boolean indicating
        if the point will be smooth for the given **segmentIndex**
        and **t**.
        """
        return self._splitAndInsertAtSegmentAndT(segmentIndex, t, False)

    def splitAndInsertPointAtSegmentAndT(self, segmentIndex, t):
        """
        Insert a point into the contour for the given
        **segmentIndex** and **t**.

        This posts a *Contour.Changed* notification.
        """
        self._splitAndInsertAtSegmentAndT(segmentIndex, t, True)

    def _splitAndInsertAtSegmentAndT(self, segmentIndex, t, insert):
        segments = self.segments
        segment = segments[segmentIndex]
        segment.insert(0, segments[segmentIndex-1][-1])
        firstPoint = segment[0]
        lastPoint = segment[-1]
        segmentType = lastPoint.segmentType
        segment = [(point.x, point.y) for point in segment]
        if segmentType == "line":
            (x1, y1), (x2, y2) = segment
            x = x1 + (x2 - x1) * t
            y = y1 + (y2 - y1) * t
            pointsToInsert = [((x, y), "line", False)]
            insertionPoint = (x, y)
            pointWillBeSmooth = False
        elif segmentType == "curve":
            pt1, pt2, pt3, pt4 = segment
            (pt1, pt2, pt3, pt4), (pt5, pt6, pt7, pt8) = bezierTools.splitCubicAtT(pt1, pt2, pt3, pt4, t)
            pointsToInsert = [(pt2, None, False), (pt3, None, False), (pt4, "curve", True), (pt6, None, False), (pt7, None, False)]
            insertionPoint = tuple(pt4)
            pointWillBeSmooth = True
        else:
            # XXX could be a quad. in that case, we could handle it.
            raise NotImplementedError("unknown segment type: %s" % segmentType)
        if insert:
            firstPointIndex = self._points.index(firstPoint)
            lastPointIndex = self._points.index(lastPoint)
            firstPoints = self._points[:firstPointIndex + 1]
            if firstPointIndex == len(self._points) - 1:
                firstPoints = firstPoints[lastPointIndex:]
                lastPoints = []
            elif lastPointIndex == 0:
                lastPoints = []
            else:
                lastPoints = self._points[lastPointIndex:]
            newPoints = [self._pointClass(pos, segmentType=segmentType, smooth=smooth) for pos, segmentType, smooth in pointsToInsert]
            self._points = firstPoints + newPoints + lastPoints
            self.dirty = True
        return insertionPoint, pointWillBeSmooth

    # -----------
    # Pen methods
    # -----------

    def beginPath(self, identifier=None):
        """
        Standard point pen *beginPath* method.
        This should not be used externally.
        """
        self.identifier = identifier

    def endPath(self):
        """
        Standard point pen *endPath* method.
        This should not be used externally.
        """
        pass

    def addPoint(self, values, segmentType=None, smooth=False, name=None, identifier=None, **kwargs):
        """
        Standard point pen *addPoint* method.
        This should not be used externally.
        """
        (x, y) = values
        point = self._pointClass((x, y), segmentType=segmentType, smooth=smooth, name=name, identifier=identifier, **kwargs)
        self.insertPoint(len(self._points), point)

    def draw(self, pen):
        """
        Draw the contour with **pen**.
        """
        from fontTools.pens.pointPen import PointToSegmentPen
        pointPen = PointToSegmentPen(pen)
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen):
        """
        Draw the contour with **pointPen**.
        """
        try:
            pointPen.beginPath(identifier=self.identifier)
        except TypeError:
            pointPen.beginPath()
            warn("The beginPath method needs an identifier kwarg. The contour's identifier value has been discarded.", DeprecationWarning)
        for point in self._points:
            try:
                pointPen.addPoint((point.x, point.y), segmentType=point.segmentType, smooth=point.smooth, name=point.name, identifier=point.identifier)
            except TypeError:
                pointPen.addPoint((point.x, point.y), segmentType=point.segmentType, smooth=point.smooth, name=point.name)
                warn("The addPoint method needs an identifier kwarg. The point's identifier value has been discarded.", DeprecationWarning)
        pointPen.endPath()

    # ----------
    # Identifier
    # ----------

    def _get_identifiers(self):
        identifiers = None
        glyph = self.glyph
        if glyph is not None:
            identifiers = glyph.identifiers
        if identifiers is None:
            identifiers = set()
        return identifiers

    identifiers = property(_get_identifiers, doc="Set of identifiers for the glyph that this contour belongs to. This is primarily for internal use.")

    def _get_identifier(self):
        return self._identifier

    def _set_identifier(self, value):
        # don't allow overwritting an existing identifier
        if self._identifier is not None:
            return
        oldIdentifier = self.identifier
        if value == oldIdentifier:
            return
        # don't allow a duplicate
        identifiers = self.identifiers
        assert value not in identifiers
        # free the old identifier
        if oldIdentifier in identifiers:
            identifiers.remove(oldIdentifier)
        # store
        self._identifier = value
        if value is not None:
            identifiers.add(value)
        # post notifications
        self.postNotification("Contour.IdentifierChanged", data=dict(oldValue=oldIdentifier, newValue=value))
        self.dirty = True

    identifier = property(_get_identifier, _set_identifier, doc="The identifier. Setting this will post *Contour.IdentifierChanged* and *Contour.Changed* notifications.")

    def generateIdentifier(self):
        """
        Create a new, unique identifier for and assign it to the contour.
        This will post *Contour.IdentifierChanged* and *Contour.Changed* notifications.
        """
        if self.identifier is None:
            identifier = makeRandomIdentifier(existing=self.identifiers)
            self.identifier = identifier
        return self.identifier

    def generateIdentifierForPoint(self, point):
        """
        Create a new, unique identifier for and assign it to the point.
        This will post *Contour.Changed* notification.
        """
        if point.identifier is None:
            identifier = makeRandomIdentifier(existing=self.identifiers)
            if identifier is not None:
                self.identifiers.add(identifier)
            point.identifier = identifier
            self.dirty = True
        return point.identifier

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(Contour, self).endSelfNotificationObservation()
        self._font = None
        self._layerSet = None
        self._layer = None
        self._glyph = None

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        def get_points(key):
            # store the point pen protocol calls
            # this will store the identifier and the point data
            pointData = []
            self.drawPoints(Recorder(pointData))
            return pointData

        getters = [('pen', get_points)]
        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        self.clear()
        self.identifier = None
        if 'pen' in data:
            # play back
            Recorder(data['pen'])(self)


class Recorder(object):
    """
    Records all method calls it receives in a list of tuples in the form of
    [(:str:command, :list:args, :dict: kwargs)]

    Method calls to be recorded must not start with an underscore.

    This class creates a callable object which can be used like a
    function: "recorder(target)" because that way calls to all methods
    that don't start with underscores can be recorded.

    This is useful to record the commands of both pen protocols
    and it may become useful for other things as well, like recording
    undo commands.

    Example Session PointPen:

    data_glyphA = []
    recorderPointPen = Recorder(data_glyphA)
    glyphA.drawPoints(recorderPointPen)

    # The point data of the glyph is now stored within data
    # we can either replay it immediately or take it away and use it
    # to replay it later

    stored_data = pickle.dumps(data_glyphA)
    restored_data_glyphA = pickle.loads(stored_data)

    player = Recorder(restored_data_glyphA)
    # The recorder behaves like glyphA.drawPoints
    player(glyphB)

    Example Session SegmentPen:

    data_glyphA = []
    recorderPen = Recorder(data_glyphA)
    glyphA.draw(recorderPen)

    # reuse it immediately
    # The recorder behaves like glyphA.draw
    recorderPen(glyphB)
    """
    def __init__(self, data=None):
        self.__dict__['_data'] = data if data is not None else []

    def __call__(self, target):
        """
        Public API.
        Replay all method calls to this Recorder to target.
        """
        for cmd, args, kwargs in self._data:
            getattr(target, cmd)(*args, **kwargs)

    def __setattr__(self, name, value):
        raise AttributeError('It\'s not allowed to set attributes here.', name)

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError(name)

        def command(*args, **kwargs):
            self._data.append((name, args, kwargs))
        # cache the method, don't use __setattr__
        self.__dict__[name] = command
        return command


if __name__ == "__main__":
    import doctest
    doctest.testmod()
